import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSPopUpButtonCell(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSPopUpNoArrow, 0)
        self.assertEqual(AppKit.NSPopUpArrowAtCenter, 1)
        self.assertEqual(AppKit.NSPopUpArrowAtBottom, 2)

        self.assertIsInstance(AppKit.NSPopUpButtonCellWillPopUpNotification, str)

    def testMethods(self):
        self.assertArgIsBOOL(AppKit.NSPopUpButtonCell.initTextCell_pullsDown_, 1)
        self.assertResultIsBOOL(AppKit.NSPopUpButtonCell.pullsDown)
        self.assertArgIsBOOL(AppKit.NSPopUpButtonCell.setPullsDown_, 0)
        self.assertResultIsBOOL(AppKit.NSPopUpButtonCell.autoenablesItems)
        self.assertArgIsBOOL(AppKit.NSPopUpButtonCell.setAutoenablesItems_, 0)
        self.assertResultIsBOOL(AppKit.NSPopUpButtonCell.usesItemFromMenu)
        self.assertArgIsBOOL(AppKit.NSPopUpButtonCell.setUsesItemFromMenu_, 0)
        self.assertResultIsBOOL(AppKit.NSPopUpButtonCell.altersStateOfSelectedItem)
        self.assertArgIsBOOL(AppKit.NSPopUpButtonCell.setAltersStateOfSelectedItem_, 0)
        self.assertResultIsBOOL(AppKit.NSPopUpButtonCell.selectItemWithTag_)



import CoreFoundation
from PyObjCTools.TestSupport import TestCase, min_os_level
import objc


class TestData(TestCase):
    def testTypes(self):
        try:
            NSCFData = objc.lookUpClass("__NSCFData")
        except objc.error:
            NSCFData = objc.lookUpClass("NSCFData")

        self.assertIs(CoreFoundation.CFDataRef, NSCFData)

    def testTypeID(self):
        v = CoreFoundation.CFDataGetTypeID()
        self.assertTrue(v, int)

    def testCreation(self):
        self.assertArgHasType(CoreFoundation.CFDataCreate, 1, b"n^v")
        self.assertArgSizeInArg(CoreFoundation.CFDataCreate, 1, 2)
        data = CoreFoundation.CFDataCreate(None, b"hello", 5)
        self.assertIsInstance(data, CoreFoundation.CFDataRef)
        bytes_data = b"hello world"
        self.assertArgHasType(CoreFoundation.CFDataCreateWithBytesNoCopy, 1, b"n^v")
        self.assertArgSizeInArg(CoreFoundation.CFDataCreateWithBytesNoCopy, 1, 2)
        data = CoreFoundation.CFDataCreateWithBytesNoCopy(
            None, bytes_data, 5, CoreFoundation.kCFAllocatorNull
        )
        self.assertIsInstance(data, CoreFoundation.CFDataRef)
        del data

        data = CoreFoundation.CFDataCreate(None, b"hello", 5)
        self.assertIsInstance(data, CoreFoundation.CFDataRef)
        cpy = CoreFoundation.CFDataCreateCopy(None, data)
        self.assertIsInstance(cpy, CoreFoundation.CFDataRef)
        cpy2 = CoreFoundation.CFDataCreateMutableCopy(None, 0, data)
        self.assertIsInstance(cpy2, CoreFoundation.CFDataRef)
        mut = CoreFoundation.CFDataCreateMutable(None, 0)
        self.assertIsInstance(mut, CoreFoundation.CFDataRef)

    def testInspection(self):
        data = CoreFoundation.CFDataCreate(None, b"hello", 5)
        self.assertIsInstance(data, CoreFoundation.CFDataRef)
        mutableData = CoreFoundation.CFDataCreateMutableCopy(None, 0, data)
        self.assertIsInstance(mutableData, CoreFoundation.CFDataRef)
        self.assertEqual(CoreFoundation.CFDataGetLength(data), 5)
        self.assertEqual(CoreFoundation.CFDataGetLength(mutableData), 5)
        v = CoreFoundation.CFDataGetBytePtr(data)
        self.assertEqual(CoreFoundation.CFDataGetBytePtr(data)[0], b"h")
        v = CoreFoundation.CFDataGetMutableBytePtr(mutableData)
        self.assertEqual(v[0], b"h")
        v[0] = b"p"

        v = CoreFoundation.CFDataGetBytePtr(mutableData)
        self.assertEqual(v[0], b"p")
        self.assertArgHasType(CoreFoundation.CFDataGetBytes, 2, b"o^v")
        self.assertArgSizeInArg(CoreFoundation.CFDataGetBytes, 2, 1)
        bytes_data = CoreFoundation.CFDataGetBytes(data, (1, 3), None)
        self.assertEqual(bytes_data, b"hello"[1:4])

        CoreFoundation.CFDataSetLength(mutableData, 3)
        self.assertEqual(CoreFoundation.CFDataGetLength(mutableData), 3)
        CoreFoundation.CFDataIncreaseLength(mutableData, 17)
        self.assertEqual(CoreFoundation.CFDataGetLength(mutableData), 20)
        CoreFoundation.CFDataSetLength(mutableData, 3)

        self.assertArgHasType(CoreFoundation.CFDataAppendBytes, 1, b"n^v")
        self.assertArgSizeInArg(CoreFoundation.CFDataAppendBytes, 1, 2)
        CoreFoundation.CFDataAppendBytes(mutableData, b" world", 6)
        self.assertEqual(CoreFoundation.CFDataGetLength(mutableData), 9)
        self.assertEqual(
            CoreFoundation.CFDataGetBytes(mutableData, (0, 9), None), b"pel world"
        )

        self.assertArgHasType(CoreFoundation.CFDataReplaceBytes, 2, b"n^v")
        self.assertArgSizeInArg(CoreFoundation.CFDataReplaceBytes, 2, 3)
        CoreFoundation.CFDataReplaceBytes(mutableData, (0, 3), b"hello", 5)
        self.assertEqual(
            CoreFoundation.CFDataGetBytes(mutableData, (0, 9), None), b"hello world"[:9]
        )

        CoreFoundation.CFDataDeleteBytes(mutableData, (0, 6))
        self.assertEqual(
            CoreFoundation.CFDataGetBytes(mutableData, (0, 5), None), b"world"
        )

    @min_os_level("10.6")
    def testConstants10_6(self):
        self.assertEqual(CoreFoundation.kCFDataSearchBackwards, 1 << 0)
        self.assertEqual(CoreFoundation.kCFDataSearchAnchored, 1 << 1)

    @min_os_level("10.6")
    def testFunctions10_6(self):
        data = CoreFoundation.CFDataCreate(None, b"hello world", 11)
        self.assertIsInstance(data, CoreFoundation.CFDataRef)
        src = CoreFoundation.CFDataCreate(None, b"wor", 3)
        self.assertIsInstance(src, CoreFoundation.CFDataRef)
        self.assertResultHasType(
            CoreFoundation.CFDataFind, CoreFoundation.CFRange.__typestr__
        )
        self.assertArgHasType(
            CoreFoundation.CFDataFind, 2, CoreFoundation.CFRange.__typestr__
        )
        v = CoreFoundation.CFDataFind(data, src, (0, 11), 0)
        self.assertIsInstance(v, CoreFoundation.CFRange)
        self.assertEqual(v, (6, 3))



import Foundation
import objc
import CoreFoundation
from PyObjCTools.TestSupport import TestCase, min_os_level

MachPortClasses = tuple(
    cls for cls in objc.getClassList() if cls.__name__ == "NSMachPort"
)


class TestMachPort(TestCase):
    def testTypes(self):
        try:
            if objc.lookUpClass("NSMachPort") is CoreFoundation.CFMachPortRef:
                return
        except objc.error:
            pass
        self.assertIsCFType(CoreFoundation.CFMachPortRef)

    def testTypeID(self):
        self.assertIsInstance(CoreFoundation.CFMachPortGetTypeID(), int)

    @min_os_level("10.8")
    def testCreate10_8(self):
        class Context:
            pass

        context = Context()

        def callout(port, msg, size, info):
            pass

        port, shouldFree = CoreFoundation.CFMachPortCreate(None, callout, context, None)

        # On OSX 10.7 or earlier this test passed, on OSX 10.8 it doesn't???
        self.assertIsInstance(port, MachPortClasses)

    def testCreate(self):
        class Context:
            pass

        context = Context()

        def callout(port, msg, size, info):
            pass

        # XXX: This one cannot be tested without bindings to the low-level mach_port API's
        # port, shouldFree = CoreFoundation.CFMachPortCreateWithPort(None, 1, callout, context, None)  # noqa: B950
        # self.assertIsInstance(port, CoreFoundation.CFMachPortRef)
        # self.assertTrue(shouldFree is True or shouldFree is False)

        port, shouldFree = CoreFoundation.CFMachPortCreate(None, callout, context, None)

        self.assertIsInstance(port, MachPortClasses)
        self.assertIsInstance(port, Foundation.NSPort)
        self.assertTrue(shouldFree is True or shouldFree is False)
        idx = CoreFoundation.CFMachPortGetPort(port)
        self.assertIsInstance(idx, int)
        ctx = CoreFoundation.CFMachPortGetContext(port, None)
        self.assertIs(ctx, context)
        cb = CoreFoundation.CFMachPortGetInvalidationCallBack(port)
        self.assertIs(cb, None)
        global didInvalidate
        didInvalidate = False

        def invalidate(port, info):
            global didInvalidate
            didInvalidate = True

        CoreFoundation.CFMachPortSetInvalidationCallBack(port, invalidate)
        cb = CoreFoundation.CFMachPortGetInvalidationCallBack(port)
        self.assertIs(invalidate, cb)
        rls = CoreFoundation.CFMachPortCreateRunLoopSource(None, port, 0)
        self.assertIsInstance(rls, CoreFoundation.CFRunLoopSourceRef)
        self.assertTrue(CoreFoundation.CFMachPortIsValid(port))
        CoreFoundation.CFMachPortInvalidate(port)
        self.assertFalse(CoreFoundation.CFMachPortIsValid(port))
        self.assertTrue(didInvalidate)



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSLayoutAnchor(TestCase):
    @min_os_level("10.12")
    def test_methods10_12(self):
        self.assertResultIsBOOL(AppKit.NSLayoutAnchor.hasAmbiguousLayout)



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSTextTable(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSTextBlockAbsoluteValueType, 0)
        self.assertEqual(AppKit.NSTextBlockPercentageValueType, 1)

        self.assertEqual(AppKit.NSTextBlockWidth, 0)
        self.assertEqual(AppKit.NSTextBlockMinimumWidth, 1)
        self.assertEqual(AppKit.NSTextBlockMaximumWidth, 2)
        self.assertEqual(AppKit.NSTextBlockHeight, 4)
        self.assertEqual(AppKit.NSTextBlockMinimumHeight, 5)
        self.assertEqual(AppKit.NSTextBlockMaximumHeight, 6)

        self.assertEqual(AppKit.NSTextBlockPadding, -1)
        self.assertEqual(AppKit.NSTextBlockBorder, 0)
        self.assertEqual(AppKit.NSTextBlockMargin, 1)

        self.assertEqual(AppKit.NSTextBlockTopAlignment, 0)
        self.assertEqual(AppKit.NSTextBlockMiddleAlignment, 1)
        self.assertEqual(AppKit.NSTextBlockBottomAlignment, 2)
        self.assertEqual(AppKit.NSTextBlockBaselineAlignment, 3)

        self.assertEqual(AppKit.NSTextTableAutomaticLayoutAlgorithm, 0)
        self.assertEqual(AppKit.NSTextTableFixedLayoutAlgorithm, 1)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSTextTable.collapsesBorders)
        self.assertArgIsBOOL(AppKit.NSTextTable.setCollapsesBorders_, 0)
        self.assertResultIsBOOL(AppKit.NSTextTable.hidesEmptyCells)
        self.assertArgIsBOOL(AppKit.NSTextTable.setHidesEmptyCells_, 0)



# import classes required to start application
import WSTApplicationDelegateClass  # noqa: F401
import WSTConnectionWindowControllerClass  # noqa: F401
from PyObjCTools import AppHelper
from twisted.internet import cfreactor

cfreactor.install()


# pass control to the AppKit
AppHelper.runEventLoop()



"""
Instances of WSTConnectionWindowController are the controlling object
for the document windows for the Web Services Tool application.

Implements a standard toolbar.
"""

import string
import sys
import traceback
import types
from threading import Thread

import Cocoa

# Note about multi-threading.
# Although WST does its network stuff in a background thread, with Python 2.2
# there are still moments where the app appears to hang briefly. This should
# only be noticeable when your DNS is slow-ish. The hang is caused by the
# socket.getaddrinfo() function, which is used (indirectly) when connecting
# to a server, which is a frequent operation when using xmlrpclib (it makes
# a new connection for each request). Up to (and including) version 2.3b1,
# Python would not grant time to other threads while blocking inside
# getaddrinfo(). This has been fixed *after* 2.3b1 was released. (jvr)
import objc

try:
    from Queue import Queue
except ImportError:
    from queue import Queue

try:
    from xmlrpclib import ServerProxy
except ImportError:
    from xmlrpc.client import ServerProxy


kWSTReloadContentsToolbarItemIdentifier = "WST: Reload Contents Toolbar Identifier"
"""Identifier for 'reload contents' toolbar item."""

kWSTPreferencesToolbarItemIdentifier = "WST: Preferences Toolbar Identifier"
"""Identifier for 'preferences' toolbar item."""

kWSTUrlTextFieldToolbarItemIdentifier = "WST: URL Textfield Toolbar Identifier"
"""Idnetifier for URL text field toolbar item."""


def addToolbarItem(
    aController,
    anIdentifier,
    aLabel,
    aPaletteLabel,
    aToolTip,
    aTarget,
    anAction,
    anItemContent,
    aMenu,
):
    """
    Adds an freshly created item to the toolbar defined by
    aController.  Makes a number of assumptions about the
    implementation of aController.  It should be refactored into a
    generically useful toolbar management untility.
    """
    toolbarItem = Cocoa.NSToolbarItem.alloc().initWithItemIdentifier_(anIdentifier)

    toolbarItem.setLabel_(aLabel)
    toolbarItem.setPaletteLabel_(aPaletteLabel)
    toolbarItem.setToolTip_(aToolTip)
    toolbarItem.setTarget_(aTarget)
    if anAction:
        toolbarItem.setAction_(anAction)

    if type(anItemContent) == Cocoa.NSImage:
        toolbarItem.setImage_(anItemContent)
    else:
        toolbarItem.setView_(anItemContent)
        bounds = anItemContent.bounds()
        minSize = (100, bounds[1][1])
        maxSize = (1000, bounds[1][1])
        toolbarItem.setMinSize_(minSize)
        toolbarItem.setMaxSize_(maxSize)

    if aMenu:
        menuItem = Cocoa.NSMenuItem.alloc().init()
        menuItem.setSubmenu_(aMenu)
        menuItem.setTitle_(aMenu.title())
        toolbarItem.setMenuFormRepresentation_(menuItem)

    aController._toolbarItems[anIdentifier] = toolbarItem


class WorkerThread(Thread):
    def __init__(self):
        """Create a worker thread. Start it by calling the start() method."""
        self.queue = Queue()
        Thread.__init__(self)

    def stop(self):
        """Stop the thread a.s.a.p., meaning whenever the currently running
        job is finished."""
        self.working = 0
        self.queue.put(None)

    def scheduleWork(self, func, *args, **kwargs):
        """Schedule some work to be done in the worker thread."""
        self.queue.put((func, args, kwargs))

    def run(self):
        """Fetch work from a queue, block when there's nothing to do.
        This method is called by Thread, don't call it yourself."""
        self.working = 1
        while self.working:
            work = self.queue.get()
            if work is None or not self.working:
                break
            func, args, kwargs = work
            pool = Cocoa.NSAutoreleasePool.alloc().init()  # noqa: F841
            try:
                func(*args, **kwargs)
            finally:
                # delete all local references; if they are the last refs they
                # may invoke autoreleases, which should then end up in our pool
                del func, args, kwargs, work
                del pool


class WSTConnectionWindowController(Cocoa.NSWindowController):
    methodDescriptionTextView = objc.IBOutlet()
    methodsTable = objc.IBOutlet()
    progressIndicator = objc.IBOutlet()
    statusTextField = objc.IBOutlet()
    urlTextField = objc.IBOutlet()

    __slots__ = (
        "_toolbarItems",
        "_toolbarDefaultItemIdentifiers",
        "_toolbarAllowedItemIdentifiers",
        "_methods",
        "_methodSignatures",
        "_methodDescriptions",
        "_server",
        "_methodPrefix",
        "_workQueue",
        "_working",
        "_workerThread",
        "_windowIsClosing",
    )

    @classmethod
    def connectionWindowController(self):
        """
        Create and return a default connection window instance.
        """
        return WSTConnectionWindowController.alloc().init()

    def init(self):
        """
        Designated initializer.

        Returns self (as per ObjC designated initializer definition,
        unlike Python's __init__() method).
        """
        self = self.initWithWindowNibName_("WSTConnection")

        self._toolbarItems = {}
        self._toolbarDefaultItemIdentifiers = []
        self._toolbarAllowedItemIdentifiers = []

        self._methods = []
        self._working = 0
        self._windowIsClosing = 0
        self._workerThread = WorkerThread()
        self._workerThread.start()
        return self

    def awakeFromNib(self):
        """
        Invoked when the NIB file is loaded.  Initializes the various
        UI widgets.
        """
        self.retain()  # balanced by autorelease() in windowWillClose_

        self.statusTextField.setStringValue_("No host specified.")
        self.progressIndicator.setStyle_(Cocoa.NSProgressIndicatorSpinningStyle)
        self.progressIndicator.setDisplayedWhenStopped_(False)

        self.createToolbar()

    def windowWillClose_(self, aNotification):
        """
        Clean up when the document window is closed.
        """
        # We must stop the worker thread and wait until it actually finishes before
        # we can allow the window to close. Weird stuff happens if we simply let the
        # thread run. When this thread is idle (blocking in queue.get()) there is
        # no problem and we can almost instantly close the window. If it's actually
        # in the middle of working it may take a couple of seconds, as we can't
        # _force_ the thread to stop: we have to ask it to to stop itself.
        self._windowIsClosing = 1  # try to stop the thread a.s.a.p.
        self._workerThread.stop()  # signal the thread that there is no more work to do
        self._workerThread.join()  # wait until it finishes
        self.autorelease()

    def createToolbar(self):
        """
        Creates and configures the toolbar to be used by the window.
        """
        toolbar = Cocoa.NSToolbar.alloc().initWithIdentifier_("WST Connection Window")
        toolbar.setDelegate_(self)
        toolbar.setAllowsUserCustomization_(True)
        toolbar.setAutosavesConfiguration_(True)

        self.createToolbarItems()

        self.window().setToolbar_(toolbar)

        lastURL = Cocoa.NSUserDefaults.standardUserDefaults().stringForKey_("LastURL")
        if lastURL and len(lastURL):
            self.urlTextField.setStringValue_(lastURL)

    def createToolbarItems(self):
        """
        Creates all of the toolbar items that can be made available in
        the toolbar.  The actual set of available toolbar items is
        determined by other mechanisms (user defaults, for example).
        """
        addToolbarItem(
            self,
            kWSTReloadContentsToolbarItemIdentifier,
            "Reload",
            "Reload",
            "Reload Contents",
            None,
            "reloadVisibleData:",
            Cocoa.NSImage.imageNamed_("Reload"),
            None,
        )
        addToolbarItem(
            self,
            kWSTPreferencesToolbarItemIdentifier,
            "Preferences",
            "Preferences",
            "Show Preferences",
            None,
            "orderFrontPreferences:",
            Cocoa.NSImage.imageNamed_("Preferences"),
            None,
        )
        addToolbarItem(
            self,
            kWSTUrlTextFieldToolbarItemIdentifier,
            "URL",
            "URL",
            "Server URL",
            None,
            None,
            self.urlTextField,
            None,
        )

        self._toolbarDefaultItemIdentifiers = [
            kWSTReloadContentsToolbarItemIdentifier,
            kWSTUrlTextFieldToolbarItemIdentifier,
            Cocoa.NSToolbarSeparatorItemIdentifier,
            Cocoa.NSToolbarCustomizeToolbarItemIdentifier,
        ]

        self._toolbarAllowedItemIdentifiers = [
            kWSTReloadContentsToolbarItemIdentifier,
            kWSTUrlTextFieldToolbarItemIdentifier,
            Cocoa.NSToolbarSeparatorItemIdentifier,
            Cocoa.NSToolbarSpaceItemIdentifier,
            Cocoa.NSToolbarFlexibleSpaceItemIdentifier,
            Cocoa.NSToolbarPrintItemIdentifier,
            kWSTPreferencesToolbarItemIdentifier,
            Cocoa.NSToolbarCustomizeToolbarItemIdentifier,
        ]

    def toolbarDefaultItemIdentifiers_(self, anIdentifier):
        """
        Return an array of toolbar item identifiers that identify the
        set, in order, of items that should be displayed on the
        default toolbar.
        """
        return self._toolbarDefaultItemIdentifiers

    def toolbarAllowedItemIdentifiers_(self, anIdentifier):
        """
        Return an array of toolbar items that may be used in the toolbar.
        """
        return self._toolbarAllowedItemIdentifiers

    def toolbar_itemForItemIdentifier_willBeInsertedIntoToolbar_(
        self, toolbar, itemIdentifier, flag
    ):
        """
        Delegate method fired when the toolbar is about to insert an
        item into the toolbar.  Item is identified by itemIdentifier.

        Effectively makes a copy of the cached reference instance of
        the toolbar item identified by itemIdentifier.
        """
        newItem = Cocoa.NSToolbarItem.alloc().initWithItemIdentifier_(itemIdentifier)
        item = self._toolbarItems[itemIdentifier]

        newItem.setLabel_(item.label())
        newItem.setPaletteLabel_(item.paletteLabel())
        if item.view():
            newItem.setView_(item.view())
        else:
            newItem.setImage_(item.image())

        newItem.setToolTip_(item.toolTip())
        newItem.setTarget_(item.target())
        newItem.setAction_(item.action())
        newItem.setMenuFormRepresentation_(item.menuFormRepresentation())

        if newItem.view():
            newItem.setMinSize_(item.minSize())
            newItem.setMaxSize_(item.maxSize())

        return newItem

    def setStatusTextFieldMessage_(self, aMessage):
        """
        Sets the contents of the statusTextField to aMessage and
        forces the fileld's contents to be redisplayed.
        """
        if not aMessage:
            aMessage = "Displaying information about %d methods." % len(self._methods)
        # All UI calls should be directed to the main thread
        self.statusTextField.performSelectorOnMainThread_withObject_waitUntilDone_(
            "setStringValue:", aMessage, 0
        )

    def reloadData(self):
        """Tell the main thread to update the table view."""
        self.methodsTable.performSelectorOnMainThread_withObject_waitUntilDone_(
            "reloadData", None, 0
        )

    def startWorking(self):
        """Signal the UI there's work going on."""
        if not self._working:
            self.progressIndicator.startAnimation_(self)
        self._working += 1

    def stopWorking(self):
        """Signal the UI that the work is done."""
        self._working -= 1
        if not self._working:
            self.progressIndicator.performSelectorOnMainThread_withObject_waitUntilDone_(
                "stopAnimation:", self, 0
            )

    @objc.IBAction
    def reloadVisibleData_(self, sender):
        """
        Reloads the list of methods and their signatures from the
        XML-RPC server specified in the urlTextField.  Displays
        appropriate error messages, if necessary.
        """
        if self._working:
            # don't start a new job while there's an unfinished one
            return
        url = self.urlTextField.stringValue()
        self._methods = []
        self._methodSignatures = {}
        self._methodDescriptions = {}

        if not url:
            self.window().setTitle_("Untitled.")
            self.setStatusTextFieldMessage_("No URL specified.")
            return

        self.window().setTitle_(url)
        Cocoa.NSUserDefaults.standardUserDefaults().setObject_forKey_(url, "LastURL")

        self.setStatusTextFieldMessage_("Retrieving method list...")
        self.startWorking()
        self._workerThread.scheduleWork(self.getMethods, url)

    @objc.python_method
    def getMethods(self, url):
        self._server = ServerProxy(url)
        pool = (
            Cocoa.NSAutoreleasePool.alloc().init()
        )  # use an extra pool to get rid of intermediates
        try:
            self._methods = self._server.listMethods()
            self._methodPrefix = ""
        except:  # noqa: E722, B001
            try:
                self._methods = self._server.system.listMethods()
                self._methodPrefix = "system."
            except:  # noqa: E722, B001
                self._server = None
                self._methodPrefix = None
                self.setStatusTextFieldMessage_(
                    "Server failed to respond to listMethods query.  "
                    "See below for more information."
                )

                exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
                self.methodDescriptionTextView.performSelectorOnMainThread_withObject_waitUntilDone_(  # noqa: B950
                    "setString:",
                    "Exception information\n\nType: %s\n\nValue: %s\n\nTraceback:\n\n %s\n"
                    % (
                        exceptionType,
                        exceptionValue,
                        "\n".join(traceback.format_tb(exceptionTraceback)),
                    ),
                    0,
                )
                self.stopWorking()
                return

        del pool
        if self._windowIsClosing:
            return

        self._methods.sort()
        self.reloadData()
        self.setStatusTextFieldMessage_(
            "Retrieving information about %d methods." % len(self._methods)
        )

        index = 0
        for aMethod in self._methods:
            if self._windowIsClosing:
                return
            pool = (
                Cocoa.NSAutoreleasePool.alloc().init()
            )  # use an extra pool to get rid of intermediates
            index = index + 1
            if not (index % 5):
                self.reloadData()
            self.setStatusTextFieldMessage_(
                "Retrieving signature for method %s (%d of %d)."
                % (aMethod, index, len(self._methods))
            )
            del pool
            methodSignature = getattr(
                self._server, self._methodPrefix + "methodSignature"
            )(aMethod)
            signatures = None
            if isinstance(methodSignature, str):
                continue
            if not len(methodSignature):
                continue

            for aSignature in methodSignature:
                if (type(aSignature) == types.ListType) and (  # noqa: E721
                    len(aSignature) > 0
                ):
                    signature = "{} {}({})".format(
                        aSignature[0],
                        aMethod,
                        string.join(aSignature[1:], ", "),
                    )
                else:
                    signature = aSignature
            if signatures:
                signatures = signatures + ", " + signature
            else:
                signatures = signature
            self._methodSignatures[aMethod] = signatures
        self.setStatusTextFieldMessage_(None)
        self.reloadData()
        self.stopWorking()

    def tableViewSelectionDidChange_(self, sender):
        """
        When the user selects a remote method, this method displays
        the documentation for that method as returned by the XML-RPC
        server.  If the method's documentation has been previously
        queried, the documentation will be retrieved from a cache.
        """
        selectedRow = self.methodsTable.selectedRow()
        selectedMethod = self._methods[selectedRow]

        if selectedMethod not in self._methodDescriptions:
            self._methodDescriptions[
                selectedMethod
            ] = "<description is being retrieved>"
            self.startWorking()

            def work():
                self.setStatusTextFieldMessage_(
                    "Retrieving signature for method %s..." % selectedMethod
                )
                methodDescription = getattr(
                    self._server, self._methodPrefix + "methodHelp"
                )(selectedMethod)
                if not methodDescription:
                    methodDescription = "No description available."
                self._methodDescriptions[selectedMethod] = methodDescription
                if selectedRow == self.methodsTable.selectedRow():
                    self.setStatusTextFieldMessage_(None)
                    self.methodDescriptionTextView.setString_(methodDescription)
                self.stopWorking()

            self._workerThread.scheduleWork(work)
        else:
            self.setStatusTextFieldMessage_(None)
        methodDescription = self._methodDescriptions[selectedMethod]
        self.methodDescriptionTextView.setString_(methodDescription)

    def numberOfRowsInTableView_(self, aTableView):
        """
        Returns the number of methods found on the server.
        """
        return len(self._methods)

    def tableView_objectValueForTableColumn_row_(
        self, aTableView, aTableColumn, rowIndex
    ):
        """
        Returns either the raw method name or the method signature,
        depending on if a signature had been found on the server.
        """
        aMethod = self._methods[rowIndex]
        if aMethod in self._methodSignatures:
            return self._methodSignatures[aMethod]
        else:
            return aMethod

    def tableView_shouldEditTableColumn_row_(self, aTableView, aTableColumn, rowIndex):
        # don't allow editing of any cells
        return 0



"""HotKey.py

An example that shows how to use Carbon HotKeys from a PyObjC application,
and how to use an NSApplication subclass.

To build the demo program, run this line in Terminal.app:

    $ python setup.py py2app -A

This creates a directory "dist" containing HotKey.app. (The
-A option causes the files to be symlinked to the .app bundle instead
of copied. This means you don't have to rebuild the app if you edit the
sources or nibs.)

NOTE: This example requires Python 2 because it uses the "Carbon"
      module that was removed in Python 3.
"""
from Carbon.CarbonEvt import GetApplicationEventTarget, RegisterEventHotKey
from Carbon.Events import cmdKey, controlKey

import Cocoa
from objc import super
from PyObjCTools import AppHelper

kEventHotKeyPressedSubtype = 6
kEventHotKeyReleasedSubtype = 9


class HotKeyApp(Cocoa.NSApplication):
    def finishLaunching(self):
        super().finishLaunching()
        # register cmd-control-J
        self.hotKeyRef = RegisterEventHotKey(
            38, cmdKey | controlKey, (0, 0), GetApplicationEventTarget(), 0
        )

    def sendEvent_(self, theEvent):
        if (
            theEvent.type() == Cocoa.NSSystemDefined
            and theEvent.subtype() == kEventHotKeyPressedSubtype
        ):
            self.activateIgnoringOtherApps_(True)
            Cocoa.NSRunAlertPanel(
                "Hot Key Pressed", "Hot Key Pressed", None, None, None
            )
        super().sendEvent_(theEvent)


if __name__ == "__main__":
    AppHelper.runEventLoop()



import struct
import sys

from PyObjCTools.TestSupport import TestCase
import objc

NSObject = objc.lookUpClass("NSObject")


class OC_TestConveniences(NSObject):
    def initWithHashValue_(self, hashValue):
        self = objc.super(OC_TestConveniences, self).init()
        self.hashValue = hashValue
        return self

    def hash(self):  # noqa: A003
        return self.hashValue


class TestConveniences(TestCase):
    def testHash(self):
        for hashValue in (0, sys.maxsize, sys.maxsize + 1, 0xFFFFFFFF):
            expect = struct.unpack("l", struct.pack("L", hashValue))[0]
            # Python can't hash to -1.  Surprise! :)
            if expect == -1:
                expect = -2
            o = OC_TestConveniences.alloc().initWithHashValue_(hashValue)
            self.assertEqual(o.hash(), hashValue)
            self.assertEqual(
                hash(o),
                expect,
                f"o.hash() == 0x{o.hash():X} | {hash(o)!r} != {expect!r}",
            )



import Foundation
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSSortDescriptor(TestCase):
    def testMethods(self):
        self.assertArgIsBOOL(Foundation.NSSortDescriptor.initWithKey_ascending_, 1)
        self.assertArgIsBOOL(
            Foundation.NSSortDescriptor.initWithKey_ascending_selector_, 1
        )
        self.assertArgIsSEL(
            Foundation.NSSortDescriptor.initWithKey_ascending_selector_, 2, b"i@:@"
        )

        self.assertResultIsBOOL(Foundation.NSSortDescriptor.ascending)

    @min_os_level("10.6")
    def testMethods10_6(self):
        self.assertArgIsBOOL(
            Foundation.NSSortDescriptor.sortDescriptorWithKey_ascending_, 1
        )
        self.assertArgIsBOOL(
            Foundation.NSSortDescriptor.sortDescriptorWithKey_ascending_selector_, 1
        )

        self.assertArgIsBOOL(
            Foundation.NSSortDescriptor.sortDescriptorWithKey_ascending_comparator_, 1
        )
        self.assertArgIsBlock(
            Foundation.NSSortDescriptor.sortDescriptorWithKey_ascending_comparator_,
            2,
            b"i@@",
        )
        self.assertArgIsBOOL(
            Foundation.NSSortDescriptor.initWithKey_ascending_comparator_, 1
        )
        self.assertArgIsBlock(
            Foundation.NSSortDescriptor.initWithKey_ascending_comparator_, 2, b"i@@"
        )
        self.assertResultIsBlock(Foundation.NSSortDescriptor.comparator, b"i@@")



"""ClassBrowser.py -- A simple class browser, demonstrating the use of an NSBrowser.

To build the demo program, run this line in Terminal.app:

    $ python setup.py py2app -A

This creates a directory "dist" containing ClassBrowser.app. (The
-A option causes the files to be symlinked to the .app bundle instead
of copied. This means you don't have to rebuild the app if you edit the
sources or nibs.)

See also the iClass demo.
"""
import sys

import objc
from Foundation import NSObject
from objc import getClassList, objc_object
from PyObjCTools import AppHelper

try:
    import AddressBook  # noqa: F401
except ImportError:
    pass

try:
    import PreferencePanes  # noqa: F401
except ImportError:
    pass

try:
    import InterfaceBuilder  # noqa: F401
except ImportError:
    pass


def _sortClasses(classList):
    classes = [(cls.__name__, cls) for cls in classList]
    classes.sort()
    return [cls for name, cls in classes]


if sys.version_info[0] == 2:

    def formatSelector(selector):
        if selector.isClassMethod:
            return "+" + selector.selector
        else:
            return "-" + selector.selector


else:

    def formatSelector(selector):
        if selector.isClassMethod:
            return "+" + selector.selector.decode("ascii")
        else:
            return "-" + selector.selector.decode("ascii")


class ClassBrowserDelegate(NSObject):
    browser = objc.IBOutlet()
    pathLabel = objc.IBOutlet()
    table = objc.IBOutlet()

    selectedClassMethods = None

    def awakeFromNib(self):
        classes = getClassList()
        rootClasses = []
        for cls in classes:
            if cls.__base__ == objc_object:
                rootClasses.append(cls)
        rootClasses = _sortClasses(rootClasses)
        self.columns = [rootClasses]
        self.browser.setMaxVisibleColumns_(7)
        self.browser.setMinColumnWidth_(150)
        self.selectedClass = None
        self.selectedClassMethods = None

    def browser_willDisplayCell_atRow_column_(self, browser, cell, row, col):
        cell.setLeaf_(not self.columns[col][row].__subclasses__())
        cell.setStringValue_(self.columns[col][row].__name__)

    def browser_numberOfRowsInColumn_(self, browser, col):
        if col == 0:
            return len(self.columns[0])
        del self.columns[col:]
        cls = self.columns[col - 1][browser.selectedRowInColumn_(col - 1)]
        subclasses = _sortClasses(cls.__subclasses__())
        self.columns.append(subclasses)
        return len(subclasses)

    @objc.IBAction
    def browserAction_(self, browser):
        self.pathLabel.setStringValue_(browser.path())

        self.selectedClass = None
        self.selectedClassMethods = None
        col = len(self.columns)
        row = -1
        while row == -1:
            col -= 1
            if col < 0:
                break
            row = self.browser.selectedRowInColumn_(col)

        if row >= 0:
            self.selectedClass = self.columns[col][row]
            self.selectedClassMethods = [
                formatSelector(obj)
                for obj in self.selectedClass.pyobjc_instanceMethods.__dict__.values()
                if hasattr(obj, "selector")
            ]
            self.selectedClassMethods += [
                formatSelector(obj)
                for obj in self.selectedClass.pyobjc_classMethods.__dict__.values()
                if hasattr(obj, "selector")
            ]
            self.selectedClassMethods.sort()

        self.table.reloadData()

    # table view delegate methods
    def numberOfRowsInTableView_(self, tableView):
        if self.selectedClassMethods is None:
            return 0
        return len(self.selectedClassMethods)

    def tableView_objectValueForTableColumn_row_(self, tableView, col, row):
        return str(self.selectedClassMethods[row])

    def tableView_shouldEditTableColumn_row_(self, tableView, col, row):
        return 0


if __name__ == "__main__":
    AppHelper.runEventLoop()



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSTreeController(TestCase):
    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSTreeController.canInsert)
        self.assertResultIsBOOL(AppKit.NSTreeController.canInsertChild)
        self.assertResultIsBOOL(AppKit.NSTreeController.canAddChild)
        self.assertResultIsBOOL(AppKit.NSTreeController.avoidsEmptySelection)
        self.assertArgIsBOOL(AppKit.NSTreeController.setAvoidsEmptySelection_, 0)
        self.assertResultIsBOOL(AppKit.NSTreeController.preservesSelection)
        self.assertArgIsBOOL(AppKit.NSTreeController.setPreservesSelection_, 0)
        self.assertResultIsBOOL(AppKit.NSTreeController.selectsInsertedObjects)
        self.assertArgIsBOOL(AppKit.NSTreeController.setSelectsInsertedObjects_, 0)
        self.assertResultIsBOOL(AppKit.NSTreeController.alwaysUsesMultipleValuesMarker)
        self.assertArgIsBOOL(
            AppKit.NSTreeController.setAlwaysUsesMultipleValuesMarker_, 0
        )
        self.assertResultIsBOOL(AppKit.NSTreeController.setSelectionIndexPaths_)
        self.assertResultIsBOOL(AppKit.NSTreeController.setSelectionIndexPath_)
        self.assertResultIsBOOL(AppKit.NSTreeController.addSelectionIndexPaths_)
        self.assertResultIsBOOL(AppKit.NSTreeController.removeSelectionIndexPaths_)



"""
Script for building the example:

Usage:
    python3 setup.py py2app
"""
from setuptools import setup

plist = {
    "CFBundleDocumentTypes": [
        {
            "CFBundleTypeExtensions": ["CurrencyConvBinding", "*"],
            "CFBundleTypeName": "CurrencyConvBinding File",
            "CFBundleTypeRole": "Editor",
            "NSDocumentClass": "CurrencyConvBindingDocument",
        }
    ]
}

setup(
    name="CurrencyConvBinding",
    app=["CurrencyConvBinding.py"],
    data_files=["English.lproj"],
    options={"py2app": {"plist": plist}},
    setup_requires=["py2app", "pyobjc-framework-Cocoa"],
)



import Foundation
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSTextCheckingResult(TestCase):
    @min_os_level("10.6")
    def testConstants(self):
        self.assertEqual(Foundation.NSTextCheckingTypeOrthography, 1 << 0)
        self.assertEqual(Foundation.NSTextCheckingTypeSpelling, 1 << 1)
        self.assertEqual(Foundation.NSTextCheckingTypeGrammar, 1 << 2)
        self.assertEqual(Foundation.NSTextCheckingTypeDate, 1 << 3)
        self.assertEqual(Foundation.NSTextCheckingTypeAddress, 1 << 4)
        self.assertEqual(Foundation.NSTextCheckingTypeLink, 1 << 5)
        self.assertEqual(Foundation.NSTextCheckingTypeQuote, 1 << 6)
        self.assertEqual(Foundation.NSTextCheckingTypeDash, 1 << 7)
        self.assertEqual(Foundation.NSTextCheckingTypeReplacement, 1 << 8)
        self.assertEqual(Foundation.NSTextCheckingTypeCorrection, 1 << 9)

        self.assertEqual(Foundation.NSTextCheckingAllSystemTypes, 0xFFFFFFFF)
        self.assertEqual(Foundation.NSTextCheckingAllCustomTypes, (0xFFFFFFFF << 32))
        self.assertEqual(
            Foundation.NSTextCheckingAllTypes,
            (
                Foundation.NSTextCheckingAllSystemTypes
                | Foundation.NSTextCheckingAllCustomTypes
            ),
        )

        self.assertIsInstance(Foundation.NSTextCheckingNameKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingJobTitleKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingOrganizationKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingStreetKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingCityKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingStateKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingZIPKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingCountryKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingPhoneKey, str)

    @min_os_level("10.7")
    def testConstants10_7(self):
        self.assertEqual(Foundation.NSTextCheckingTypeRegularExpression, 1 << 10)
        self.assertEqual(Foundation.NSTextCheckingTypePhoneNumber, 1 << 11)
        self.assertEqual(Foundation.NSTextCheckingTypeTransitInformation, 1 << 12)

        self.assertIsInstance(Foundation.NSTextCheckingAirlineKey, str)
        self.assertIsInstance(Foundation.NSTextCheckingFlightKey, str)

    @min_os_level("10.6")
    def testMethods(self):
        self.assertResultHasType(
            Foundation.NSTextCheckingResult.range, Foundation.NSRange.__typestr__
        )

        self.assertArgHasType(
            Foundation.NSTextCheckingResult.orthographyCheckingResultWithRange_orthography_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.spellCheckingResultWithRange_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.grammarCheckingResultWithRange_details_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.dateCheckingResultWithRange_date_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.dateCheckingResultWithRange_date_timeZone_duration_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.addressCheckingResultWithRange_components_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.linkCheckingResultWithRange_URL_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.quoteCheckingResultWithRange_replacementString_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.dashCheckingResultWithRange_replacementString_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.replacementCheckingResultWithRange_replacementString_,  # noqa: B950
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.correctionCheckingResultWithRange_replacementString_,  # noqa: B950
            0,
            Foundation.NSRange.__typestr__,
        )

    @min_os_level("10.7")
    def testMethods10_7(self):
        self.assertArgHasType(
            Foundation.NSTextCheckingResult.regularExpressionCheckingResultWithRanges_count_regularExpression_,  # noqa: B950
            0,
            b"n^" + Foundation.NSRange.__typestr__,
        )
        self.assertArgSizeInArg(
            Foundation.NSTextCheckingResult.regularExpressionCheckingResultWithRanges_count_regularExpression_,  # noqa: B950
            0,
            1,
        )



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSHelpManager(TestCase):
    def testConstants(self):
        self.assertIsInstance(AppKit.NSContextHelpModeDidActivateNotification, str)
        self.assertIsInstance(AppKit.NSContextHelpModeDidDeactivateNotification, str)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSHelpManager.isContextHelpModeActive)
        self.assertArgIsBOOL(AppKit.NSHelpManager.setContextHelpModeActive_, 0)
        self.assertResultIsBOOL(
            AppKit.NSHelpManager.showContextHelpForObject_locationHint_
        )

    @min_os_level("10.11")
    def testMethods10_6(self):
        self.assertResultIsBOOL(AppKit.NSHelpManager.registerBooksInBundle_)



#!/usr/bin/env python
"""
This script using NSNetServiceBrowser to look for local HTTP servers.
"""

import objc
from Foundation import NSDate, NSNetServiceBrowser, NSObject, NSRunLoop

objc.setVerbose(1)


class PrintingResolverDelegate(NSObject):
    def netServiceDidResolveAddress_(self, service):
        addresses = service.addresses()
        if len(addresses) == 0:
            return
        print(f"{service.name()}.{service.domain()}")
        for address in service.addresses():
            print(f"   {address}")
        print("")
        service.setDelegate_(None)

    def netService_didNotResolve_(self, service, didNotResolve):
        print("didNotResolve", didNotResolve)
        service.setDelegate_(None)


class PrintingBrowserDelegate(NSObject):
    def startLookup(self):
        self.delegates = []
        for aNetService in self.services:
            prd = PrintingResolverDelegate.new()
            aNetService.setDelegate_(prd)
            aNetService.resolve()
            self.delegates.append(prd)

    def netServiceBrowserWillSearch_(self, browser):
        print("Browsing for advertised services...")
        self.services = []

    def netServiceBrowserDidStopSearch_(self, browser):
        print("Browse complete")
        self.startLookup()

    def netServiceBrowser_didNotSearch_(self, browser, errorDict):
        print("Could not search.")

    def netServiceBrowser_didFindService_moreComing_(
        self, browser, aNetService, moreComing
    ):
        print(f"Found a service: {aNetService.name()} {aNetService.domain()}")
        self.services.append(aNetService)
        if not moreComing:
            browser.stop()

    def netServiceBrowser_didRemoveService_moreComing_(
        self, browser, aNetService, moreComing
    ):
        print(f"Service removed: {aNetService.name()}")
        if not moreComing:
            browser.stop()


def findDomains(serviceName, seconds=5.0):
    runloop = NSRunLoop.currentRunLoop()
    browser = NSNetServiceBrowser.new()
    pbd = PrintingBrowserDelegate.new()
    browser.setDelegate_(pbd)
    browser.searchForServicesOfType_inDomain_(serviceName, "")
    untilWhen = NSDate.dateWithTimeIntervalSinceNow_(seconds)
    runloop.runUntilDate_(untilWhen)


if __name__ == "__main__":
    # Use '_afpovertcp' instead of '_http' to look for fileservers.
    findDomains("_afpovertcp._tcp")



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSImageRep(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSImageRepMatchesDevice, 0)

        self.assertEqual(
            AppKit.NSImageRepRegistryChangedNotification,
            AppKit.NSImageRepRegistryDidChangeNotification,
        )
        self.assertIsInstance(AppKit.NSImageRepRegistryDidChangeNotification, str)

        self.assertEqual(AppKit.NSImageLayoutDirectionUnspecified, -1)
        self.assertEqual(AppKit.NSImageLayoutDirectionLeftToRight, 2)
        self.assertEqual(AppKit.NSImageLayoutDirectionRightToLeft, 3)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSImageRep.draw)
        self.assertResultIsBOOL(AppKit.NSImageRep.drawAtPoint_)
        self.assertResultIsBOOL(AppKit.NSImageRep.drawInRect_)
        self.assertArgIsBOOL(AppKit.NSImageRep.setAlpha_, 0)
        self.assertResultIsBOOL(AppKit.NSImageRep.hasAlpha)
        self.assertArgIsBOOL(AppKit.NSImageRep.setOpaque_, 0)
        self.assertResultIsBOOL(AppKit.NSImageRep.isOpaque)
        self.assertResultIsBOOL(AppKit.NSImageRep.canInitWithData_)
        self.assertResultIsBOOL(AppKit.NSImageRep.canInitWithPasteboard_)

    @min_os_level("10.6")
    def testMethods10_6(self):
        self.assertResultIsBOOL(
            AppKit.NSImageRep.drawInRect_fromRect_operation_fraction_respectFlipped_hints_
        )
        self.assertArgHasType(
            AppKit.NSImageRep.drawInRect_fromRect_operation_fraction_respectFlipped_hints_,
            0,
            AppKit.NSRect.__typestr__,
        )
        self.assertArgHasType(
            AppKit.NSImageRep.drawInRect_fromRect_operation_fraction_respectFlipped_hints_,
            1,
            AppKit.NSRect.__typestr__,
        )
        self.assertArgIsBOOL(
            AppKit.NSImageRep.drawInRect_fromRect_operation_fraction_respectFlipped_hints_,
            4,
        )

        self.assertArgHasType(
            AppKit.NSImageRep.CGImageForProposedRect_context_hints_,
            0,
            b"N^" + AppKit.NSRect.__typestr__,
        )



import CoreFoundation
from PyObjCTools.TestSupport import TestCase, expectedFailure


class TestXMLParser(TestCase):
    # Note: This doesn't actually test the API

    def testTypes(self):
        self.assertIsCFType(CoreFoundation.CFXMLParserRef)

    def testTypeID(self):
        self.assertIsInstance(CoreFoundation.CFXMLParserGetTypeID(), int)

    def testConstants(self):
        self.assertEqual(CoreFoundation.kCFXMLParserValidateDocument, (1 << 0))
        self.assertEqual(CoreFoundation.kCFXMLParserSkipMetaData, (1 << 1))
        self.assertEqual(CoreFoundation.kCFXMLParserReplacePhysicalEntities, (1 << 2))
        self.assertEqual(CoreFoundation.kCFXMLParserSkipWhitespace, (1 << 3))
        self.assertEqual(CoreFoundation.kCFXMLParserResolveExternalEntities, (1 << 4))
        self.assertEqual(CoreFoundation.kCFXMLParserAddImpliedAttributes, (1 << 5))
        self.assertEqual(CoreFoundation.kCFXMLParserAllOptions, 0x00FFFFFF)
        self.assertEqual(CoreFoundation.kCFXMLParserNoOptions, 0)
        self.assertEqual(CoreFoundation.kCFXMLStatusParseNotBegun, -2)
        self.assertEqual(CoreFoundation.kCFXMLStatusParseInProgress, -1)
        self.assertEqual(CoreFoundation.kCFXMLStatusParseSuccessful, 0)
        self.assertEqual(CoreFoundation.kCFXMLErrorUnexpectedEOF, 1)
        self.assertEqual(CoreFoundation.kCFXMLErrorUnknownEncoding, 2)
        self.assertEqual(CoreFoundation.kCFXMLErrorEncodingConversionFailure, 3)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedProcessingInstruction, 4)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedDTD, 5)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedName, 6)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedCDSect, 7)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedCloseTag, 8)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedStartTag, 9)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedDocument, 10)
        self.assertEqual(CoreFoundation.kCFXMLErrorElementlessDocument, 11)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedComment, 12)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedCharacterReference, 13)
        self.assertEqual(CoreFoundation.kCFXMLErrorMalformedParsedCharacterData, 14)
        self.assertEqual(CoreFoundation.kCFXMLErrorNoData, 15)
        self.assertIsInstance(CoreFoundation.kCFXMLTreeErrorDescription, str)
        self.assertIsInstance(CoreFoundation.kCFXMLTreeErrorLineNumber, str)
        self.assertIsInstance(CoreFoundation.kCFXMLTreeErrorLocation, str)
        self.assertIsInstance(CoreFoundation.kCFXMLTreeErrorStatusCode, str)

    @expectedFailure
    def testMissingWrappers(self):
        self.fail("Missing manual wrappers for CoreFoundation.CFXMLParser (low prio)")



import Foundation
from PyObjCTools.TestSupport import TestCase


class TestNSScriptCommandDescriptionHelper(Foundation.NSScriptCommandDescription):
    def isOptionalArgumentWithName_(self, name):
        return 1


class TestNSScriptCommandDescription(TestCase):
    def testMethods(self):
        # This should be tested on the actual Foundation.NSScriptCommandDescription class,
        # but for some reason that class doesn't seem to have the
        # required method (at least not without instantating the class)
        self.assertResultIsBOOL(
            TestNSScriptCommandDescriptionHelper.isOptionalArgumentWithName_
        )



import Foundation
import objc
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSExceptionInteraction(TestCase):
    def testRepeatedAllocInit(self):
        for _ in range(1, 1000):
            _ = Foundation.NSException.alloc().initWithName_reason_userInfo_(
                "Bogus", "A bad reason", {"foo": "bar"}
            )

    def testFormat(self):
        try:
            Foundation.NSException.raise_format_(
                "ExceptionName", "Format: %s %d", b"hello", 42
            )

        except TypeError:
            raise

        except objc.error as e:
            self.assertEqual(e._pyobjc_info_["name"], "ExceptionName")
            self.assertEqual(e._pyobjc_info_["reason"], "Format: hello 42")


class TestNSException(TestCase):
    def testConstants(self):
        self.assertIsInstance(Foundation.NSGenericException, str)
        self.assertIsInstance(Foundation.NSRangeException, str)
        self.assertIsInstance(Foundation.NSInvalidArgumentException, str)
        self.assertIsInstance(Foundation.NSInternalInconsistencyException, str)
        self.assertIsInstance(Foundation.NSMallocException, str)
        self.assertIsInstance(Foundation.NSObjectInaccessibleException, str)
        self.assertIsInstance(Foundation.NSObjectNotAvailableException, str)
        self.assertIsInstance(Foundation.NSDestinationInvalidException, str)
        self.assertIsInstance(Foundation.NSPortTimeoutException, str)
        self.assertIsInstance(Foundation.NSInvalidSendPortException, str)
        self.assertIsInstance(Foundation.NSInvalidReceivePortException, str)
        self.assertIsInstance(Foundation.NSPortSendException, str)
        self.assertIsInstance(Foundation.NSPortReceiveException, str)
        self.assertIsInstance(Foundation.NSOldStyleException, str)

    @min_os_level("10.6")
    def testConstants10_6(self):
        self.assertIsInstance(Foundation.NSAssertionHandlerKey, str)

    def testUncaughtExceptionHandler(self):
        self.assertArgIsFunction(
            Foundation.NSSetUncaughtExceptionHandler, 0, b"v@", True
        )
        self.assertResultIsFunction(Foundation.NSGetUncaughtExceptionHandler, b"v@")

    def testNoAssert(self):
        self.assertNotHasAttr(Foundation, "NSAssert5")
        self.assertNotHasAttr(Foundation, "NSAssert4")
        self.assertNotHasAttr(Foundation, "NSAssert3")
        self.assertNotHasAttr(Foundation, "NSAssert2")
        self.assertNotHasAttr(Foundation, "NSAssert1")
        self.assertNotHasAttr(Foundation, "NSAssert")
        self.assertNotHasAttr(Foundation, "NSParameterAssert")
        self.assertNotHasAttr(Foundation, "NSCAssert5")
        self.assertNotHasAttr(Foundation, "NSCAssert4")
        self.assertNotHasAttr(Foundation, "NSCAssert3")
        self.assertNotHasAttr(Foundation, "NSCAssert2")
        self.assertNotHasAttr(Foundation, "NSCAssert1")
        self.assertNotHasAttr(Foundation, "NSCAssert")
        self.assertNotHasAttr(Foundation, "NSCParameterAssert")

    def testMethods(self):
        self.assertArgIsPrintf(Foundation.NSException.raise_format_, 1)

        self.assertArgIsPrintf(
            Foundation.NSAssertionHandler.handleFailureInMethod_object_file_lineNumber_description_,  # noqa: B950
            4,
        )
        self.assertArgIsPrintf(
            Foundation.NSAssertionHandler.handleFailureInFunction_file_lineNumber_description_,
            3,
        )



import CoreFoundation
from PyObjCTools.TestSupport import TestCase


class TestStringTokenizer(TestCase):
    def testTypes(self):
        self.assertIsCFType(CoreFoundation.CFStringTokenizerRef)

    def testFunctions(self):
        s = "Spring eens over een boom"
        v = CoreFoundation.CFStringTokenizerCopyBestStringLanguage(s, (0, len(s)))
        self.assertEqual(v, "nl")

        v = CoreFoundation.CFStringTokenizerGetTypeID()
        self.assertIsInstance(v, int)
        self.assertResultIsCFRetained(CoreFoundation.CFStringTokenizerCreate)
        tok = CoreFoundation.CFStringTokenizerCreate(
            None,
            s,
            (0, len(s)),
            CoreFoundation.kCFStringTokenizerUnitWord,
            CoreFoundation.CFLocaleCopyCurrent(),
        )
        self.assertIsInstance(tok, CoreFoundation.CFStringTokenizerRef)
        v = CoreFoundation.CFStringTokenizerGoToTokenAtIndex(tok, 2)
        self.assertEqual(v, CoreFoundation.kCFStringTokenizerTokenNormal)

        v = CoreFoundation.CFStringTokenizerGetCurrentTokenRange(tok)
        self.assertEqual(v, CoreFoundation.CFRange(0, 6))

        v = CoreFoundation.CFStringTokenizerAdvanceToNextToken(tok)
        self.assertEqual(v, CoreFoundation.kCFStringTokenizerTokenNormal)

        v = CoreFoundation.CFStringTokenizerGetCurrentTokenRange(tok)
        self.assertEqual(v, CoreFoundation.CFRange(7, 4))

        self.assertResultIsCFRetained(
            CoreFoundation.CFStringTokenizerCopyCurrentTokenAttribute
        )
        v = CoreFoundation.CFStringTokenizerCopyCurrentTokenAttribute(
            tok, CoreFoundation.kCFStringTokenizerAttributeLanguage
        )
        # self.assertIs(v, None)

        s = "A dog jumped over a log. And then some more."
        CoreFoundation.CFStringTokenizerSetString(tok, s, (0, len(s)))

        subref = []
        idx, ranges = CoreFoundation.CFStringTokenizerGetCurrentSubTokens(
            tok, None, 20, subref
        )
        self.assertEqual(idx, 0)
        self.assertEqual(ranges, ())
        self.assertEqual(subref, [])

    def testConstants(self):
        self.assertEqual(CoreFoundation.kCFStringTokenizerUnitWord, 0)
        self.assertEqual(CoreFoundation.kCFStringTokenizerUnitSentence, 1)
        self.assertEqual(CoreFoundation.kCFStringTokenizerUnitParagraph, 2)
        self.assertEqual(CoreFoundation.kCFStringTokenizerUnitLineBreak, 3)
        self.assertEqual(CoreFoundation.kCFStringTokenizerUnitWordBoundary, 4)
        self.assertEqual(
            CoreFoundation.kCFStringTokenizerAttributeLatinTranscription, 1 << 16
        )
        self.assertEqual(CoreFoundation.kCFStringTokenizerAttributeLanguage, 1 << 17)
        self.assertEqual(CoreFoundation.kCFStringTokenizerTokenNone, 0)
        self.assertEqual(CoreFoundation.kCFStringTokenizerTokenNormal, 1)
        self.assertEqual(CoreFoundation.kCFStringTokenizerTokenHasSubTokensMask, 1 << 1)
        self.assertEqual(
            CoreFoundation.kCFStringTokenizerTokenHasDerivedSubTokensMask, 1 << 2
        )
        self.assertEqual(
            CoreFoundation.kCFStringTokenizerTokenHasHasNumbersMask, 1 << 3
        )
        self.assertEqual(
            CoreFoundation.kCFStringTokenizerTokenHasNonLettersMask, 1 << 4
        )
        self.assertEqual(CoreFoundation.kCFStringTokenizerTokenIsCJWordMask, 1 << 5)



import Foundation
from PyObjCTools.TestSupport import TestCase, min_os_level
import objc


class TestNSIndexSet(TestCase):
    def testConvenience(self):
        v = Foundation.NSIndexSet.indexSetWithIndexesInRange_((5, 10))
        lst = list(v)
        self.assertEqual(lst, [5, 6, 7, 8, 9, 10, 11, 12, 13, 14])

        lst = list(reversed(v))
        self.assertEqual(lst, [14, 13, 12, 11, 10, 9, 8, 7, 6, 5])

        v2 = Foundation.NSIndexSet.indexSetWithIndexesInRange_((5, 9))
        v3 = Foundation.NSIndexSet.indexSetWithIndexesInRange_((5, 10))

        self.assertFalse(v == v2)
        self.assertTrue(v == v3)
        self.assertFalse(v == 42)

        self.assertTrue(v != v2)
        self.assertTrue(v != 42)
        self.assertFalse(v != v3)

        self.assertTrue(8 in v)
        self.assertFalse(16 in v)
        self.assertFalse("a" in v)

    def testMethods(self):
        self.assertResultIsBOOL(Foundation.NSIndexSet.isEqualToIndexSet_)
        self.assertResultIsBOOL(Foundation.NSIndexSet.containsIndex_)
        self.assertResultIsBOOL(Foundation.NSIndexSet.containsIndexesInRange_)
        self.assertResultIsBOOL(Foundation.NSIndexSet.containsIndexes_)
        self.assertResultIsBOOL(Foundation.NSIndexSet.intersectsIndexesInRange_)

        self.assertArgIsOut(Foundation.NSIndexSet.getIndexes_maxCount_inIndexRange_, 0)
        self.assertArgSizeInArg(
            Foundation.NSIndexSet.getIndexes_maxCount_inIndexRange_, 0, 1
        )
        self.assertArgSizeInResult(
            Foundation.NSIndexSet.getIndexes_maxCount_inIndexRange_, 0
        )

    @min_os_level("10.6")
    def testMethods10_6(self):
        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateIndexesUsingBlock_,
            0,
            b"v" + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateIndexesWithOptions_usingBlock_,
            1,
            b"v" + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgHasType(
            Foundation.NSIndexSet.enumerateIndexesInRange_options_usingBlock_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateIndexesInRange_options_usingBlock_,
            2,
            b"v" + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )

        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexPassingTest_,
            0,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexWithOptions_passingTest_,
            1,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgHasType(
            Foundation.NSIndexSet.indexInRange_options_passingTest_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexInRange_options_passingTest_,
            2,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )

        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexesPassingTest_,
            0,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexesWithOptions_passingTest_,
            1,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )
        self.assertArgHasType(
            Foundation.NSIndexSet.indexesInRange_options_passingTest_,
            0,
            Foundation.NSRange.__typestr__,
        )
        self.assertArgIsBlock(
            Foundation.NSIndexSet.indexesInRange_options_passingTest_,
            2,
            objc._C_NSBOOL + objc._C_NSUInteger + b"o^" + objc._C_NSBOOL,
        )

    @min_os_level("10.7")
    def testMethod10_7(self):
        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateRangesUsingBlock_,
            0,
            b"v" + Foundation.NSRange.__typestr__ + b"o^" + objc._C_NSBOOL,
        )

        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateRangesWithOptions_usingBlock_,
            1,
            b"v" + Foundation.NSRange.__typestr__ + b"o^" + objc._C_NSBOOL,
        )

        self.assertArgIsBlock(
            Foundation.NSIndexSet.enumerateRangesInRange_options_usingBlock_,
            2,
            b"v" + Foundation.NSRange.__typestr__ + b"o^" + objc._C_NSBOOL,
        )



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level, min_sdk_level
import objc


class TestNSTextContent(TestCase):
    @min_os_level("11.0")
    def test_constants11_0(self):
        self.assertIsInstance(AppKit.NSTextContentTypeUsername, str)
        self.assertIsInstance(AppKit.NSTextContentTypePassword, str)
        self.assertIsInstance(AppKit.NSTextContentTypeOneTimeCode, str)

    @min_sdk_level("11.0")
    def test_protocols(self):
        objc.protocolNamed("NSTextContent")



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSColorList(TestCase):
    def testConstants(self):
        self.assertIsInstance(AppKit.NSColorListDidChangeNotification, str)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSColorList.isEditable)
        self.assertResultIsBOOL(AppKit.NSColorList.writeToFile_)

    @min_os_level("10.11")
    def testMethods10_11(self):
        self.assertResultIsBOOL(AppKit.NSColorList.writeToURL_error_)
        self.assertArgIsOut(AppKit.NSColorList.writeToURL_error_, 1)



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSImage(TestCase):
    def testMethods(self):
        self.assertArgIsBOOL(AppKit.NSImageView.setEditable_, 0)
        self.assertResultIsBOOL(AppKit.NSImageView.isEditable)
        self.assertArgIsBOOL(AppKit.NSImageView.setAnimates_, 0)
        self.assertResultIsBOOL(AppKit.NSImageView.animates)
        self.assertArgIsBOOL(AppKit.NSImageView.setAllowsCutCopyPaste_, 0)
        self.assertResultIsBOOL(AppKit.NSImageView.allowsCutCopyPaste)



import BookmarksDocument  # noqa: F401
import DNDArrayController  # noqa: F401
import DNDTableView  # noqa: F401

if __name__ == "__main__":
    from PyObjCTools import AppHelper

    AppHelper.runEventLoop()



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSParagraphStyle(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSLeftTabStopType, 0)
        self.assertEqual(AppKit.NSRightTabStopType, 1)
        self.assertEqual(AppKit.NSCenterTabStopType, 2)
        self.assertEqual(AppKit.NSDecimalTabStopType, 3)

        self.assertEqual(AppKit.NSLineBreakByWordWrapping, 0)
        self.assertEqual(AppKit.NSLineBreakByCharWrapping, 1)
        self.assertEqual(AppKit.NSLineBreakByClipping, 2)
        self.assertEqual(AppKit.NSLineBreakByTruncatingHead, 3)
        self.assertEqual(AppKit.NSLineBreakByTruncatingTail, 4)
        self.assertEqual(AppKit.NSLineBreakByTruncatingMiddle, 5)

        self.assertIsInstance(AppKit.NSTabColumnTerminatorsAttributeName, str)

        self.assertEqual(AppKit.NSLineBreakStrategyNone, 0)
        self.assertEqual(AppKit.NSLineBreakStrategyPushOut, 1 << 0)
        self.assertEqual(AppKit.NSLineBreakStrategyHangulWordPriority, 1 << 1)
        self.assertEqual(AppKit.NSLineBreakStrategyStandard, 0xFFFF)

    @min_os_level("10.11")
    def testMethods10_11(self):
        self.assertResultIsBOOL(
            AppKit.NSParagraphStyle.allowsDefaultTighteningForTruncation
        )
        self.assertArgIsBOOL(
            AppKit.NSMutableParagraphStyle.setAllowsDefaultTighteningForTruncation_, 0
        )

    @min_os_level("12.0")
    def testMethods12_0(self):
        self.assertResultIsBOOL(AppKit.NSParagraphStyle.usesDefaultHyphenation)
        self.assertArgIsBOOL(
            AppKit.NSMutableParagraphStyle.setUsesDefaultHyphenation_, 0
        )



import CoreFoundation
import objc
from PyObjCTools.TestSupport import TestCase


class TestSet(TestCase):
    def testTypes(self):
        try:
            if objc.lookUpClass("NSSet") is CoreFoundation.CFSetRef:
                return
        except objc.error:
            pass
        self.assertIsCFType(CoreFoundation.CFSetRef)

    def testTypeID(self):
        self.assertIsInstance(CoreFoundation.CFSetGetTypeID(), int)

    def testCreation(self):
        st = CoreFoundation.CFSetCreate(
            None, ["a", "b", "c"], 3, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        self.assertIsInstance(st, objc.lookUpClass("NSSet"))
        st = CoreFoundation.CFSetCreate(
            None, ["a", "b", "c"], 3, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        st = CoreFoundation.CFSetCreateMutable(
            None, 0, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        cp = CoreFoundation.CFSetCreateMutableCopy(None, 0, st)
        self.assertIsInstance(cp, CoreFoundation.CFSetRef)
        cp = CoreFoundation.CFSetCreateCopy(None, st)
        self.assertIsInstance(cp, CoreFoundation.CFSetRef)

    def testInspection(self):
        st = CoreFoundation.CFSetCreate(
            None, ["a", "b", "c"], 3, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        self.assertIsInstance(st, objc.lookUpClass("NSSet"))
        v = CoreFoundation.CFSetGetCount(st)
        self.assertEqual(v, 3)
        self.assertArgHasType(CoreFoundation.CFSetGetCountOfValue, 1, b"@")
        v = CoreFoundation.CFSetGetCountOfValue(st, "d")
        self.assertEqual(v, 0)
        v = CoreFoundation.CFSetGetCountOfValue(st, "b")
        self.assertEqual(v, 1)
        self.assertArgHasType(CoreFoundation.CFSetContainsValue, 1, b"@")
        v = CoreFoundation.CFSetContainsValue(st, "d")
        self.assertIs(v, False)
        v = CoreFoundation.CFSetContainsValue(st, "b")
        self.assertIs(v, True)
        self.assertResultHasType(CoreFoundation.CFSetGetValue, b"@")
        self.assertArgHasType(CoreFoundation.CFSetGetValue, 1, b"@")
        v = CoreFoundation.CFSetGetValue(st, "d")
        self.assertIs(v, None)
        v = CoreFoundation.CFSetGetValue(st, "b")
        self.assertEqual(v, "b")
        self.assertResultIsBOOL(CoreFoundation.CFSetGetValueIfPresent)
        self.assertArgHasType(CoreFoundation.CFSetGetValueIfPresent, 1, b"@")
        self.assertArgHasType(CoreFoundation.CFSetGetValueIfPresent, 2, b"o^@")
        present, value = CoreFoundation.CFSetGetValueIfPresent(st, "c", None)
        self.assertIs(present, True)
        self.assertEqual(value, "c")
        values = CoreFoundation.CFSetGetValues(st, None)
        values = list(values)
        values.sort()
        self.assertEqual(values, ["a", "b", "c"])

    def testApplying(self):
        st = CoreFoundation.CFSetCreate(
            None, ["a", "b", "c"], 3, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        self.assertIsInstance(st, objc.lookUpClass("NSSet"))
        context = []

        def callback(value, context):
            context.append(value)

        self.assertArgIsFunction(CoreFoundation.CFSetApplyFunction, 1, b"v@@", False)
        self.assertArgHasType(CoreFoundation.CFSetApplyFunction, 2, b"@")
        CoreFoundation.CFSetApplyFunction(st, callback, context)
        self.assertEqual(len(context), 3)
        context.sort()
        self.assertEqual(context, ["a", "b", "c"])

    def testMutation(self):
        st = CoreFoundation.CFSetCreate(
            None, ["a", "b", "c"], 3, CoreFoundation.kCFTypeSetCallBacks
        )
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        self.assertIsInstance(st, objc.lookUpClass("NSSet"))
        st = CoreFoundation.CFSetCreateMutableCopy(None, 0, st)
        self.assertIsInstance(st, CoreFoundation.CFSetRef)
        self.assertEqual(CoreFoundation.CFSetGetCount(st), 3)
        self.assertArgHasType(CoreFoundation.CFSetSetValue, 1, b"@")
        CoreFoundation.CFSetSetValue(st, "c")
        self.assertEqual(CoreFoundation.CFSetGetCount(st), 3)
        CoreFoundation.CFSetSetValue(st, "d")
        self.assertEqual(CoreFoundation.CFSetGetCount(st), 4)
        self.assertArgHasType(CoreFoundation.CFSetRemoveValue, 1, b"@")
        CoreFoundation.CFSetRemoveValue(st, "c")
        self.assertEqual(CoreFoundation.CFSetGetCount(st), 3)
        CoreFoundation.CFSetRemoveAllValues(st)
        self.assertEqual(CoreFoundation.CFSetGetCount(st), 0)
        self.assertArgHasType(CoreFoundation.CFSetAddValue, 1, b"@")
        CoreFoundation.CFSetAddValue(st, "d")

        self.assertArgHasType(CoreFoundation.CFSetReplaceValue, 1, b"@")
        CoreFoundation.CFSetReplaceValue(st, "d")



"""
Python mapping for the Cocoa framework.

This module does not contain docstrings for the wrapped code, check Apple's
documentation for details on how to use these functions and classes.
"""
import sys

import AppKit
import Foundation
import objc

mod = objc.ObjCLazyModule(
    "Cocoa",
    None,
    None,
    {},
    None,
    {
        "__doc__": __doc__,
        "objc": objc,
        "__path__": __path__,
        "__loader__": globals().get("__loader__", None),
    },
    (AppKit, Foundation),
)
sys.modules["Cocoa"] = mod



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSDraggingItem(TestCase):
    @min_os_level("10.7")
    def testConstants10_7(self):
        self.assertIsInstance(AppKit.NSDraggingImageComponentIconKey, str)
        self.assertIsInstance(AppKit.NSDraggingImageComponentLabelKey, str)

    @min_os_level("10.7")
    def testMethods10_7(self):
        self.assertArgIsBlock(
            AppKit.NSDraggingItem.setImageComponentsProvider_, 0, b"@"
        )
        self.assertResultIsBlock(AppKit.NSDraggingItem.imageComponentsProvider, b"@")



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSApplicationScriptingHelper(AppKit.NSObject):
    def application_delegateHandlesKey_(self, app, key):
        return 1


class TestNSApplicationScripting(TestCase):
    def testMethods(self):
        self.assertResultIsBOOL(
            TestNSApplicationScriptingHelper.application_delegateHandlesKey_
        )



#
#  DNDArrayController.py
#  Bookmarks
#
#  Converted by u.fiedler on 10.02.05.
#
#  The original version was written in Objective-C by Malcolm Crawford
#  at http://homepage.mac.com/mmalc/CocoaExamples/controllers.html
#
#  See "Dragging Files" for a conceptual introduction:
#  file:///Developer/ADC%20Reference%20Library/documentation/Cocoa/Conceptual/DragandDrop/index.html#//apple_ref/doc/uid/10000069i  # noqa: B950
#  or http://developer.apple.com/documentation/Cocoa/Conceptual/DragandDrop/Tasks/DraggingFiles.html  # noqa: B950

import objc
from BookmarksDocument import CopiedRowsType
from Cocoa import (
    NSURL,
    NSArrayController,
    NSCalendarDate,
    NSDragOperationCopy,
    NSDragOperationMove,
    NSIndexSet,
    NSMakeRange,
    NSMutableIndexSet,
    NSNotFound,
    NSTableViewDropAbove,
    NSURLPboardType,
)

MovedRowsType = "MOVED_ROWS_TYPE"


class DNDArrayController(NSArrayController):
    # DNDArrayController is delegate and dataSource of tableView
    tableView = objc.IBOutlet()

    def awakeFromNib(self):
        """register for drag and drop"""
        self.tableView.registerForDraggedTypes_(
            [CopiedRowsType, MovedRowsType, NSURLPboardType]
        )
        self.tableView.setAllowsMultipleSelection_(True)

    def tableView_writeRows_toPasteboard_(self, tv, rows, pboard):
        # declare our own pasteboard types
        typesArray = [CopiedRowsType, MovedRowsType]

        # If the number of rows is not 1, then we only support our own types.
        # If there is just one row, then try to create an NSURL from the url
        # value in that row.  If that's possible, add NSURLPboardType to the
        # list of supported types, and add the NSURL to the pasteboard.
        if len(rows) != 1:
            pboard.declareTypes_owner_(typesArray, self)
        else:
            # Try to create an URL
            # If we can, add NSURLPboardType to the declared types and write
            # the URL to the pasteboard; otherwise declare existing types
            row = rows[0]
            urlString = self.arrangedObjects()[row].valueForKey_("url")
            url = None
            if urlString:
                url = NSURL.URLWithString_(urlString)
            if urlString and url:
                typesArray.append(NSURLPboardType)
                pboard.declareTypes_owner_(typesArray, self)
                url.writeToPasteboard_(pboard)
            else:
                pboard.declareTypes_owner_(typesArray, self)

        # add rows array for local move
        pboard.setPropertyList_forType_(rows, MovedRowsType)

        # create new array of selected rows for remote drop
        # could do deferred provision, but keep it direct for clarity
        rowCopies = self.arrangedObjects()[:]

        # setPropertyList works here because we're using dictionaries, strings,
        # and dates; otherwise, archive collection to NSData...
        pboard.setPropertyList_forType_(rowCopies, CopiedRowsType)
        return True

    def tableView_validateDrop_proposedRow_proposedDropOperation_(
        self, tv, info, row, op
    ):
        dragOp = NSDragOperationCopy
        # if drag source is self, it's a move
        if info.draggingSource() == self.tableView:
            dragOp = NSDragOperationMove
        # we want to put the object at, not over,
        # the current row (contrast NSTableViewDropOn)
        tv.setDropRow_dropOperation_(row, NSTableViewDropAbove)
        return dragOp

    def tableView_acceptDrop_row_dropOperation_(self, tv, info, row, op):
        if row < 0:
            row = 0
        if info.draggingSource() == self.tableView:
            rows = info.draggingPasteboard().propertyListForType_(MovedRowsType)
            indexSet = self.indexSetFromRows_(rows)
            self.moveObjectsInArrangedObjectsFromIndexes_toIndex_(indexSet, row)
            # set selected rows to those that were just moved
            # Need to work out what moved where to determine proper selection...
            rowsAbove = self.rowsAboveRow_inIndexSet_(row, indexSet)
            aRange = NSMakeRange(row - rowsAbove, indexSet.count())
            indexSet = NSIndexSet.indexSetWithIndexesInRange_(aRange)
            # set selected rows to those that were just copied
            self.setSelectionIndexes_(indexSet)
            return True

        # Can we get rows from another document?  If so, add them, then return.
        newRows = info.draggingPasteboard().propertyListForType_(CopiedRowsType)
        if newRows:
            aRange = NSMakeRange(row, newRows.count())
            indexSet = NSIndexSet.indexSetWithIndexesInRange_(aRange)
            self.insertObjects_atArrangedObjectIndexes_(newRows, indexSet)
            self.setSelectionIndexes_(indexSet)
            return True

        # Can we get an URL?  If so, add a new row, configure it, then return.
        url = NSURL.URLFromPasteboard_(info.draggingPasteboard())
        if url:
            newObject = self.newObject()
            self.insertObject_atArrangedObjectIndex_(newObject, row)
            newObject.setValue_forKey_(url.absoluteString(), "url")
            newObject.setValue_forKey_(NSCalendarDate.date(), "date")
            # set selected rows to those that were just copied
            self.setSelectionIndex_(row)
            return True
        return False

    def moveObjectsInArrangedObjectsFromIndexes_toIndex_(self, indexSet, insertIndex):
        objects = self.arrangedObjects()
        index = indexSet.lastIndex()
        aboveInsertIndexCount = 0
        removeIndex = 0

        while index != NSNotFound:
            if index >= insertIndex:
                removeIndex = index + aboveInsertIndexCount
                aboveInsertIndexCount += 1
            else:
                removeIndex = index
                insertIndex -= 1
            obj = objects.objectAtIndex_(removeIndex)
            self.removeObjectAtArrangedObjectIndex_(removeIndex)
            self.insertObject_atArrangedObjectIndex_(obj, insertIndex)
            index = indexSet.indexLessThanIndex_(index)

    def indexSetFromRows_(self, rows):
        indexSet = NSMutableIndexSet.indexSet()
        for row in rows:
            indexSet.addIndex_(row)
        return indexSet

    def rowsAboveRow_inIndexSet_(self, row, indexSet):
        currentIndex = indexSet.firstIndex()
        i = 0
        while currentIndex != NSNotFound:
            if currentIndex < row:
                i += 1
            currentIndex = indexSet.indexGreaterThanIndex_(currentIndex)
        return i



import uuid

import Foundation
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSUUID(TestCase):
    @min_os_level("10.8")
    def testMethods10_8(self):
        val = Foundation.NSUUID.UUID()

        self.assertArgIsOut(val.getUUIDBytes_, 0)
        self.assertArgIsIn(val.initWithUUIDBytes_, 0)

        v = val.getUUIDBytes_(None)
        self.assertIsInstance(v, bytes)
        self.assertEqual(len(v), 16)

        w = uuid.UUID(bytes=v)

        self.assertEqual(str(w).lower(), val.UUIDString().lower())

        w = uuid.uuid1()
        val = Foundation.NSUUID.alloc().initWithUUIDBytes_(w.bytes)
        self.assertEqual(str(w).lower(), val.UUIDString().lower())



"""TableModel.py -- Minimal example showing how to use an NSTableView.

To build the demo program, run this line in Terminal.app:

    $ python setup.py py2app -A

This creates a directory "dist" containing TableModel.app. (The
-A option causes the files to be symlinked to the .app bundle instead
of copied. This means you don't have to rebuild the app if you edit the
sources or nibs.)
"""

# Or rather, this app shows how to create an object conforming to the
# NSTableDataSource protocol.  The model object is connected to the table view
# in the nib, both as the delegate and the data source.  The table view will
# ask our object for data; it controls us, not the other way around.

import objc
from Cocoa import NSObject
from PyObjCTools import AppHelper

ROWCOUNT = 200


class TableModel(NSObject):
    label = objc.IBOutlet()
    tableView = objc.IBOutlet()

    # An instance of this class is connected to the 'datasource' and
    # the 'delegate' outlet of the table view in the nib.

    def awakeFromNib(self):
        self.editedFields = {}
        self.rowcount = ROWCOUNT
        # tableView is an outlet set in Interface Builder
        self.tableView.setAutosaveName_("TableView")
        self.tableView.setAutosaveTableColumns_(1)
        self.tableView.setTarget_(self)
        self.tableView.setDoubleAction_("doubleClick:")
        self.tableView.window().setDelegate_(self)

    def init(self):
        self.rowcount = ROWCOUNT
        return self

    @objc.IBAction
    def doubleClick_(self, sender):
        # double-clicking a column header causes clickedRow() to return -1
        print("doubleClick!", sender.clickedColumn(), sender.clickedRow())

    # data source methods
    def numberOfRowsInTableView_(self, aTableView):
        return self.rowcount

    def tableView_objectValueForTableColumn_row_(
        self, aTableView, aTableColumn, rowIndex
    ):
        col = aTableColumn.identifier()
        return self.editedFields.get(
            (aTableColumn, rowIndex), "{%s, %d}" % (col, rowIndex)
        )

    def tableView_setObjectValue_forTableColumn_row_(
        self, aTableView, anObject, aTableColumn, rowIndex
    ):
        self.editedFields[(aTableColumn, rowIndex)] = anObject

    # delegate methods
    def tableView_shouldSelectRow_(self, aTableView, rowIndex):
        # only allow odd rows to be selected
        return rowIndex % 2

    def tableView_shouldEditTableColumn_row_(self, aTableView, aTableColumn, rowIndex):
        # only allow cells in the second column in odd rows to be edited
        return (rowIndex % 2) and aTableColumn.identifier() == "col_2"

    def tableViewSelectionDidChange_(self, notification):
        # Always use this delegate method instead of using the action to do something
        # when the selection changed: the action method is only called when the selection
        # changed by means of a mouse click; this method is also called when the
        # user uses the arrow keys.
        if self.tableView.numberOfSelectedRows() == 0:
            if self.tableView.numberOfSelectedColumns() == 0:
                self.label.setStringValue_("")
            else:
                items = list(self.tableView.selectedColumnEnumerator())
                word = "Column"
        else:
            items = list(self.tableView.selectedRowEnumerator())
            word = "Row"
        label = "{}{}: {}".format(
            word,
            ("s", "")[len(items) == 1],
            ", ".join([str(x) for x in items]),
        )
        self.label.setStringValue_(label)

    def windowShouldClose_(self, sender):
        print("Should Close?")
        return 0


AppHelper.runEventLoop()



import Foundation
from PyObjCTools.TestSupport import TestCase


class TestNSCompoundPredicate(TestCase):
    def testConstants(self):
        self.assertEqual(Foundation.NSNotPredicateType, 0)
        self.assertEqual(Foundation.NSAndPredicateType, 1)
        self.assertEqual(Foundation.NSOrPredicateType, 2)



import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSStringDrawing(TestCase):
    @min_os_level("10.5")
    def testConstants10_5(self):
        self.assertEqual(AppKit.NSStringDrawingTruncatesLastVisibleLine, (1 << 5))

    def testConstants(self):
        self.assertEqual(AppKit.NSStringDrawingUsesLineFragmentOrigin, (1 << 0))
        self.assertEqual(AppKit.NSStringDrawingUsesFontLeading, (1 << 1))
        self.assertEqual(AppKit.NSStringDrawingDisableScreenFontSubstitution, (1 << 2))
        self.assertEqual(AppKit.NSStringDrawingUsesDeviceMetrics, (1 << 3))
        self.assertEqual(AppKit.NSStringDrawingOneShot, (1 << 4))



import Foundation
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSInflectionRule(TestCase):
    @min_os_level("12.0")
    def test_methods(self):
        self.assertResultIsBOOL(
            Foundation.NSInflectionRule.canInflectPreferredLocalization
        )



import Foundation
import objc
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestHelper(Foundation.NSObject):
    def incFoo_(self, foo):
        foo[0] += 1


class TestNSUndoManager(TestCase):
    def testUndoManager(self):
        x = TestHelper.new()
        m = Foundation.NSUndoManager.new()
        lst = [0]

        m.prepareWithInvocationTarget_(x).incFoo_(lst)
        m.undo()

        self.assertEqual(lst[0], 1)

    def __del__(self, objc=objc):
        objc.recycleAutoreleasePool()


# Undo Integer test
# From David Eppstein
#
# test ability of int argument to pass through undo and then
# be used as parameter to another routine expecting an int
#
# the actual routine I want to use is
# Foundation.NSTableView.editColumn_row_withEvent_select_
# but that involves setting up a UI; instead use Foundation.NSIndexSpecifier

if hasattr(Foundation, "NSIndexSpecifier"):

    class TestUndoInt(TestCase):
        class UndoInt(Foundation.NSObject):
            undo = Foundation.NSUndoManager.alloc().init()
            idx = Foundation.NSIndexSpecifier.alloc().init()
            idx.setIndex_(0)

            def test_(self, i):
                self.undo.prepareWithInvocationTarget_(self).test_(self.idx.index())
                self.idx.setIndex_(i)

        def testUndoInt(self):
            # test that undo works
            x = TestUndoInt.UndoInt.alloc().init()
            x.test_(3)
            assert x.idx.index() == 3
            x.undo.undo()
            assert x.idx.index() == 0


# end Undo Integer test


class TestSubclassingUndo(TestCase):
    # Bugreport: 678759 Subclassing Foundation.NSUndoManager fails

    def testSubclass(self):
        class UndoSubclass(Foundation.NSUndoManager):
            pass

        x = TestHelper.new()
        m = UndoSubclass.new()
        lst = [0]

        m.prepareWithInvocationTarget_(x).incFoo_(lst)
        m.undo()

        self.assertEqual(lst[0], 1)

    def testConstants(self):
        self.assertIsInstance(Foundation.NSUndoManagerCheckpointNotification, str)
        self.assertIsInstance(Foundation.NSUndoManagerWillUndoChangeNotification, str)
        self.assertIsInstance(Foundation.NSUndoManagerWillRedoChangeNotification, str)
        self.assertIsInstance(Foundation.NSUndoManagerDidUndoChangeNotification, str)
        self.assertIsInstance(Foundation.NSUndoManagerDidRedoChangeNotification, str)
        self.assertIsInstance(Foundation.NSUndoManagerDidOpenUndoGroupNotification, str)
        self.assertIsInstance(
            Foundation.NSUndoManagerWillCloseUndoGroupNotification, str
        )
        self.assertEqual(Foundation.NSUndoCloseGroupingRunLoopOrdering, 350_000)

    @min_os_level("10.7")
    def testConstants10_7(self):
        self.assertIsInstance(Foundation.NSUndoManagerGroupIsDiscardableKey, str)
        self.assertIsInstance(
            Foundation.NSUndoManagerDidCloseUndoGroupNotification, str
        )

    def testMethods(self):
        self.assertResultIsBOOL(Foundation.NSUndoManager.isUndoRegistrationEnabled)
        self.assertResultIsBOOL(Foundation.NSUndoManager.groupsByEvent)
        self.assertArgIsBOOL(Foundation.NSUndoManager.setGroupsByEvent_, 0)
        self.assertResultIsBOOL(Foundation.NSUndoManager.canUndo)
        self.assertResultIsBOOL(Foundation.NSUndoManager.canRedo)
        self.assertResultIsBOOL(Foundation.NSUndoManager.isUndoing)
        self.assertResultIsBOOL(Foundation.NSUndoManager.isRedoing)
        self.assertArgIsSEL(
            Foundation.NSUndoManager.registerUndoWithTarget_selector_object_, 1, b"v@:@"
        )

    @min_os_level("10.7")
    def testMethods10_7(self):
        self.assertArgIsBOOL(Foundation.NSUndoManager.setActionIsDiscardable_, 0)
        self.assertResultIsBOOL(Foundation.NSUndoManager.undoActionIsDiscardable)
        self.assertResultIsBOOL(Foundation.NSUndoManager.redoActionIsDiscardable)

    @min_os_level("10.11")
    def testMethods10_11(self):
        self.assertArgIsBlock(
            Foundation.NSUndoManager.registerUndoWithTarget_handler_, 1, b"v@"
        )



import objc
import Foundation
import AppKit
from PyObjCTools.TestSupport import TestCase, min_os_level


class TestNSAttributedString(TestCase):
    def testMethodsFoundation(self):
        self.assertArgIsOut(
            AppKit.NSAttributedString.attributesAtIndex_effectiveRange_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.attributesAtIndex_longestEffectiveRange_inRange_,
            1,
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.attribute_atIndex_longestEffectiveRange_inRange_,
            2,
        )

        self.assertResultIsBOOL(AppKit.NSAttributedString.isEqualToAttributedString_)

    def testConstantsAppKit(self):
        self.assertIsInstance(AppKit.NSManagerDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSFontAttributeName, str)
        self.assertIsInstance(AppKit.NSParagraphStyleAttributeName, str)
        self.assertIsInstance(AppKit.NSForegroundColorAttributeName, str)
        self.assertIsInstance(AppKit.NSUnderlineStyleAttributeName, str)
        self.assertIsInstance(AppKit.NSSuperscriptAttributeName, str)
        self.assertIsInstance(AppKit.NSBackgroundColorAttributeName, str)
        self.assertIsInstance(AppKit.NSAttachmentAttributeName, str)
        self.assertIsInstance(AppKit.NSLigatureAttributeName, str)
        self.assertIsInstance(AppKit.NSBaselineOffsetAttributeName, str)
        self.assertIsInstance(AppKit.NSKernAttributeName, str)
        self.assertIsInstance(AppKit.NSLinkAttributeName, str)
        self.assertIsInstance(AppKit.NSStrokeWidthAttributeName, str)
        self.assertIsInstance(AppKit.NSStrokeColorAttributeName, str)
        self.assertIsInstance(AppKit.NSUnderlineColorAttributeName, str)
        self.assertIsInstance(AppKit.NSStrikethroughStyleAttributeName, str)
        self.assertIsInstance(AppKit.NSStrikethroughColorAttributeName, str)
        self.assertIsInstance(AppKit.NSShadowAttributeName, str)
        self.assertIsInstance(AppKit.NSObliquenessAttributeName, str)
        self.assertIsInstance(AppKit.NSExpansionAttributeName, str)
        self.assertIsInstance(AppKit.NSCursorAttributeName, str)
        self.assertIsInstance(AppKit.NSToolTipAttributeName, str)
        self.assertIsInstance(AppKit.NSCharacterShapeAttributeName, str)
        self.assertIsInstance(AppKit.NSGlyphInfoAttributeName, str)
        self.assertIsInstance(AppKit.NSMarkedClauseSegmentAttributeName, str)
        self.assertIsInstance(AppKit.NSSpellingStateAttributeName, str)

        self.assertEqual(AppKit.NSUnderlineStyleNone, 0x00)
        self.assertEqual(AppKit.NSUnderlineStyleSingle, 0x01)
        self.assertEqual(AppKit.NSUnderlineStyleThick, 0x02)
        self.assertEqual(AppKit.NSUnderlineStyleDouble, 0x09)
        self.assertEqual(AppKit.NSUnderlinePatternSolid, 0x0000)
        self.assertEqual(AppKit.NSUnderlinePatternDot, 0x0100)
        self.assertEqual(AppKit.NSUnderlinePatternDash, 0x0200)
        self.assertEqual(AppKit.NSUnderlinePatternDashDot, 0x0300)
        self.assertEqual(AppKit.NSUnderlinePatternDashDotDot, 0x0400)

        self.assertIsInstance(AppKit.NSUnderlineByWordMask, int)

        self.assertEqual(AppKit.NSSpellingStateSpellingFlag, 1)
        self.assertEqual(AppKit.NSSpellingStateGrammarFlag, 2)

        self.assertIsInstance(AppKit.NSPlainTextDocumentType, str)
        self.assertIsInstance(AppKit.NSRTFTextDocumentType, str)
        self.assertIsInstance(AppKit.NSRTFDTextDocumentType, str)
        self.assertIsInstance(AppKit.NSMacSimpleTextDocumentType, str)
        self.assertIsInstance(AppKit.NSHTMLTextDocumentType, str)
        self.assertIsInstance(AppKit.NSDocFormatTextDocumentType, str)
        self.assertIsInstance(AppKit.NSWordMLTextDocumentType, str)
        self.assertIsInstance(AppKit.NSWebArchiveTextDocumentType, str)
        self.assertIsInstance(AppKit.NSOfficeOpenXMLTextDocumentType, str)
        self.assertIsInstance(AppKit.NSOpenDocumentTextDocumentType, str)
        self.assertIsInstance(AppKit.NSPaperSizeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSLeftMarginDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSRightMarginDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSTopMarginDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSBottomMarginDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSViewSizeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSViewZoomDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSViewModeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSDocumentTypeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSReadOnlyDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSConvertedDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCocoaVersionDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSBackgroundColorDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSHyphenationFactorDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSDefaultTabIntervalDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCharacterEncodingDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSTitleDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCompanyDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCopyrightDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSSubjectDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSAuthorDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSKeywordsDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCommentDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSEditorDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCreationTimeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSModificationTimeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSExcludedElementsDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSTextEncodingNameDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSPrefixSpacesDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSDocumentTypeDocumentOption, str)
        self.assertIsInstance(AppKit.NSDefaultAttributesDocumentOption, str)
        self.assertIsInstance(AppKit.NSCharacterEncodingDocumentOption, str)
        self.assertIsInstance(AppKit.NSTextEncodingNameDocumentOption, str)
        self.assertIsInstance(AppKit.NSBaseURLDocumentOption, str)
        self.assertIsInstance(AppKit.NSTimeoutDocumentOption, str)
        self.assertIsInstance(AppKit.NSWebPreferencesDocumentOption, str)
        self.assertIsInstance(AppKit.NSWebResourceLoadDelegateDocumentOption, str)
        self.assertIsInstance(AppKit.NSTextSizeMultiplierDocumentOption, str)

        self.assertEqual(AppKit.NSNoUnderlineStyle, 0)
        self.assertEqual(AppKit.NSSingleUnderlineStyle, 1)
        self.assertIsInstance(AppKit.NSUnderlineStrikethroughMask, int)

    def testMethodsAppKit(self):
        self.assertResultIsBOOL(AppKit.NSAttributedString.containsAttachments)
        self.assertArgIsBOOL(AppKit.NSAttributedString.nextWordFromIndex_forward_, 1)
        self.assertArgIsOut(AppKit.NSAttributedString.URLAtIndex_effectiveRange_, 1)

        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithURL_options_documentAttributes_error_, 2
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithURL_options_documentAttributes_error_, 3
        )

        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithData_options_documentAttributes_error_, 2
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithData_options_documentAttributes_error_, 3
        )

        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithPath_documentAttributes_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithURL_documentAttributes_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithRTF_documentAttributes_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithRTFD_documentAttributes_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithHTML_documentAttributes_, 1
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithHTML_options_documentAttributes_, 2
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithHTML_baseURL_documentAttributes_, 2
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.initWithRTFDFileWrapper_documentAttributes_, 1
        )

        self.assertArgIsOut(
            AppKit.NSAttributedString.dataFromRange_documentAttributes_error_, 2
        )
        self.assertArgIsOut(
            AppKit.NSAttributedString.fileWrapperFromRange_documentAttributes_error_, 2
        )

        self.assertResultIsBOOL(
            AppKit.NSMutableAttributedString.readFromURL_options_documentAttributes_error_
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromURL_options_documentAttributes_error_,
            2,
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromURL_options_documentAttributes_error_,
            3,
        )

        self.assertResultIsBOOL(
            AppKit.NSMutableAttributedString.readFromData_options_documentAttributes_error_
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromData_options_documentAttributes_error_,
            2,
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromData_options_documentAttributes_error_,
            3,
        )

        self.assertResultIsBOOL(
            AppKit.NSMutableAttributedString.readFromURL_options_documentAttributes_
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromURL_options_documentAttributes_, 2
        )

        self.assertResultIsBOOL(
            AppKit.NSMutableAttributedString.readFromData_options_documentAttributes_
        )
        self.assertArgIsOut(
            AppKit.NSMutableAttributedString.readFromData_options_documentAttributes_, 2
        )

    @min_os_level("10.6")
    def testConstants10_6(self):
        self.assertEqual(AppKit.NSAttributedStringEnumerationReverse, 1 << 1)
        self.assertEqual(
            AppKit.NSAttributedStringEnumerationLongestEffectiveRangeNotRequired,
            1 << 20,
        )

        self.assertIsInstance(AppKit.NSWritingDirectionAttributeName, str)
        self.assertIsInstance(AppKit.NSFileTypeDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSCategoryDocumentAttribute, str)
        self.assertIsInstance(AppKit.NSFileTypeDocumentOption, str)

    @min_os_level("10.7")
    def testConstants10_7(self):
        self.assertIsInstance(AppKit.NSVerticalGlyphFormAttributeName, str)
        self.assertIsInstance(AppKit.NSTextLayoutSectionOrientation, str)
        self.assertIsInstance(AppKit.NSTextLayoutSectionRange, str)
        self.assertIsInstance(AppKit.NSTextLayoutSectionsAttribute, str)

    @min_os_level("10.8")
    def testConstants10_8(self):
        self.assertIsInstance(AppKit.NSTextAlternativesAttributeName, str)
        self.assertIsInstance(AppKit.NSUsesScreenFontsDocumentAttribute, str)

    @min_os_level("10.10")
    def testConstants10_10(self):
        self.assertIsInstance(AppKit.NSTextEffectAttributeName, str)
        self.assertIsInstance(AppKit.NSTextEffectLetterpressStyle, str)

    @min_os_level("12.0")
    def test_constants12_0(self):
        self.assertEqual(Foundation.NSInlinePresentationIntentEmphasized, 1 << 0)
        self.assertEqual(
            Foundation.NSInlinePresentationIntentStronglyEmphasized, 1 << 1
        )
        self.assertEqual(Foundation.NSInlinePresentationIntentCode, 1 << 2)
        self.assertEqual(Foundation.NSInlinePresentationIntentStrikethrough, 1 << 5)
        self.assertEqual(Foundation.NSInlinePresentationIntentSoftBreak, 1 << 6)
        self.assertEqual(Foundation.NSInlinePresentationIntentLineBreak, 1 << 7)
        self.assertEqual(Foundation.NSInlinePresentationIntentInlineHTML, 1 << 8)
        self.assertEqual(Foundation.NSInlinePresentationIntentBlockHTML, 1 << 9)

        self.assertIsInstance(Foundation.NSInlinePresentationIntentAttributeName, str)
        self.assertIsInstance(Foundation.NSAlternateDescriptionAttributeName, str)
        self.assertIsInstance(Foundation.NSImageURLAttributeName, str)
        self.assertIsInstance(Foundation.NSLanguageIdentifierAttributeName, str)

        self.assertEqual(
            Foundation.NSAttributedStringMarkdownParsingFailureReturnError, 0
        )
        self.assertEqual(
            Foundation.NSAttributedStringMarkdownParsingFailureReturnPartiallyParsedIfPossible,
            1,
        )

        self.assertEqual(Foundation.NSAttributedStringMarkdownInterpretedSyntaxFull, 0)
        self.assertEqual(
            Foundation.NSAttributedStringMarkdownInterpretedSyntaxInlineOnly, 1
        )

        self.assertEqual(
            Foundation.NSAttributedStringFormattingInsertArgumentAttributesWithoutMerging,
            1 << 0,
        )
        self.assertEqual(
            Foundation.NSAttributedStringFormattingApplyReplacementIndexAttribute,
            1 << 1,
        )

        self.assertIsInstance(Foundation.NSReplacementIndexAttributeName, str)
        self.assertIsInstance(Foundation.NSMorphologyAttributeName, str)
        self.assertIsInstance(Foundation.NSInflectionRuleAttributeName, str)
        self.assertIsInstance(Foundation.NSPresentationIntentAttributeName, str)
        self.assertIsInstance(Foundation.NSInflectionAlternativeAttributeName, str)

        self.assertEqual(Foundation.NSPresentationIntentKindParagraph, 0)
        self.assertEqual(Foundation.NSPresentationIntentKindHeader, 1)
        self.assertEqual(Foundation.NSPresentationIntentKindOrderedList, 2)
        self.assertEqual(Foundation.NSPresentationIntentKindUnorderedList, 3)
        self.assertEqual(Foundation.NSPresentationIntentKindListItem, 4)
        self.assertEqual(Foundation.NSPresentationIntentKindCodeBlock, 5)
        self.assertEqual(Foundation.NSPresentationIntentKindBlockQuote, 6)
        self.assertEqual(Foundation.NSPresentationIntentKindThematicBreak, 7)
        self.assertEqual(Foundation.NSPresentationIntentKindTable, 8)
        self.assertEqual(Foundation.NSPresentationIntentKindTableHeaderRow, 9)
        self.assertEqual(Foundation.NSPresentationIntentKindTableRow, 10)
        self.assertEqual(Foundation.NSPresentationIntentKindTableCell, 11)

        self.assertEqual(Foundation.NSPresentationIntentTableColumnAlignmentLeft, 0)
        self.assertEqual(Foundation.NSPresentationIntentTableColumnAlignmentCenter, 1)
        self.assertEqual(Foundation.NSPresentationIntentTableColumnAlignmentRight, 2)

    @min_os_level("10.6")
    def testMethods10_6(self):
        self.assertArgHasType(
            AppKit.NSAttributedString.enumerateAttributesInRange_options_usingBlock_,
            0,
            AppKit.NSRange.__typestr__,
        )
        self.assertArgIsBlock(
            AppKit.NSAttributedString.enumerateAttributesInRange_options_usingBlock_,
            2,
            b"v@" + AppKit.NSRange.__typestr__ + b"o^" + objc._C_NSBOOL,
        )

        self.assertArgHasType(
            AppKit.NSAttributedString.enumerateAttribute_inRange_options_usingBlock_,
            1,
            AppKit.NSRange.__typestr__,
        )
        self.assertArgIsBlock(
            AppKit.NSAttributedString.enumerateAttribute_inRange_options_usingBlock_,
            3,
            b"v@" + AppKit.NSRange.__typestr__ + b"o^" + objc._C_NSBOOL,
        )

    @min_os_level("12.0")
    def test_methods12_0(self):
        self.assertResultIsBOOL(
            Foundation.NSAttributedStringMarkdownParsingOptions.allowsExtendedAttributes
        )
        self.assertArgIsBOOL(
            Foundation.NSAttributedStringMarkdownParsingOptions.setAllowsExtendedAttributes_,
            0,
        )

        self.assertArgIsOut(
            Foundation.NSAttributedString.initWithContentsOfMarkdownFileAtURL_options_baseURL_error_,
            3,
        )
        self.assertArgIsOut(
            Foundation.NSAttributedString.initWithMarkdown_options_baseURL_error_, 3
        )
        self.assertArgIsOut(
            Foundation.NSAttributedString.initWithMarkdownString_options_baseURL_error_,
            3,
        )

        self.assertArgIsPrintf(
            Foundation.NSAttributedString.initWithFormat_options_locale_, 0
        )
        self.assertArgIsPrintf(
            Foundation.NSAttributedString.localizedAttributedStringWithFormat_, 0
        )
        self.assertArgIsPrintf(
            Foundation.NSAttributedString.localizedAttributedStringWithFormat_options_,
            0,
        )

        self.assertArgIsPrintf(
            Foundation.NSMutableAttributedString.appendLocalizedFormat_, 0
        )

        self.assertResultIsBOOL(
            Foundation.NSPresentationIntent.isEquivalentToPresentationIntent_
        )



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSGradient(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSGradientDrawsBeforeStartingLocation, (1 << 0))
        self.assertEqual(AppKit.NSGradientDrawsAfterEndingLocation, (1 << 1))

    def testMethods(self):
        self.assertArgSizeInArg(
            AppKit.NSGradient.initWithColors_atLocations_colorSpace_, 1, 0
        )
        self.assertArgIsIn(AppKit.NSGradient.initWithColors_atLocations_colorSpace_, 1)

        self.assertArgIsOut(AppKit.NSGradient.getColor_location_atIndex_, 0)
        self.assertArgIsOut(AppKit.NSGradient.getColor_location_atIndex_, 1)



import array

import AppKit
from PyObjCTools.TestSupport import TestCase


class TestRegressions(TestCase):
    def testQualifiersInSignature(self):
        AppKit.NSColor.redColor().getRed_green_blue_alpha_(None, None, None, None)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSColor.ignoresAlpha)
        self.assertArgIsBOOL(AppKit.NSColor.setIgnoresAlpha_, 0)

        space = AppKit.NSColorSpace.adobeRGB1998ColorSpace()
        color = AppKit.NSColor.colorWithColorSpace_components_count_(
            space, (0.1, 0.2, 0.3, 0.4), 4
        )
        self.assertIsInstance(color, AppKit.NSColor)

        color = AppKit.NSColor.colorWithCalibratedRed_green_blue_alpha_(0, 0, 0, 0)
        r, g, b, a = color.getRed_green_blue_alpha_(None, None, None, None)
        self.assertIsInstance(r, float)
        self.assertIsInstance(g, float)
        self.assertIsInstance(b, float)
        self.assertIsInstance(a, float)

        color = AppKit.NSColor.colorWithCalibratedHue_saturation_brightness_alpha_(
            0.1, 0.2, 0.3, 0.4
        )
        h, s, b, a = color.getHue_saturation_brightness_alpha_(None, None, None, None)
        self.assertIsInstance(h, float)
        self.assertIsInstance(s, float)
        self.assertIsInstance(b, float)
        self.assertIsInstance(a, float)

        color = AppKit.NSColor.colorWithCalibratedWhite_alpha_(0.1, 0.2)
        w, a = color.getWhite_alpha_(None, None)
        self.assertIsInstance(w, float)
        self.assertIsInstance(a, float)

        color = AppKit.NSColor.colorWithDeviceCyan_magenta_yellow_black_alpha_(
            1, 1, 1, 1, 1
        )
        c, m, y, b, a = color.getCyan_magenta_yellow_black_alpha_(
            None, None, None, None, None
        )
        self.assertIsInstance(c, float)
        self.assertIsInstance(m, float)
        self.assertIsInstance(y, float)
        self.assertIsInstance(b, float)
        self.assertIsInstance(a, float)

        a = array.array("d", [0] * 6)
        v = color.getComponents_(a)
        self.assertIs(v, a)
        self.assertEqual(a[0], 1.0)

    def testConstants(self):
        self.assertIsInstance(AppKit.NSSystemColorsDidChangeNotification, str)

        self.assertEqual(AppKit.NSAppKitVersionNumberWithPatternColorLeakFix, 641.0)

        self.assertEqual(AppKit.NSColorTypeComponentBased, 0)
        self.assertEqual(AppKit.NSColorTypePattern, 1)
        self.assertEqual(AppKit.NSColorTypeCatalog, 2)

        self.assertEqual(AppKit.NSColorSystemEffectNone, 0)
        self.assertEqual(AppKit.NSColorSystemEffectPressed, 1)
        self.assertEqual(AppKit.NSColorSystemEffectDeepPressed, 2)
        self.assertEqual(AppKit.NSColorSystemEffectDisabled, 3)
        self.assertEqual(AppKit.NSColorSystemEffectRollover, 4)



"""
Tests for the Key-Value Coding for hybrid python objects.

NOTE: Testcases here should be synchronized with the Key-Value Coding tests
in PyObjCTools.test.test_keyvalue and objc.test.test_keyvalue.

TODO:
    - This test uses C code, that code should be added to this package!
    - Tests that access properties in the parent Objective-C class!
    - More key-error tests, the tests don't cover all relevant code yet.
"""
import Foundation
from PyObjCTest.testhelper import PyObjC_TestClass3 as STUB
from PyObjCTools.TestSupport import TestCase, max_os_level
import objc


class KeyValueClass1(Foundation.NSObject):
    def init(self):
        self = objc.super(KeyValueClass1, self).init()
        self.key3 = 3
        self._key4 = "4"
        self.__private = "private"
        return self

    def addMultiple(self):
        self.multiple = KeyValueClass1.alloc().init()
        self.multiple.level2 = KeyValueClass1.alloc().init()
        self.multiple.level2.level3 = KeyValueClass1.alloc().init()
        self.multiple.level2.level3.keyA = "hello"
        self.multiple.level2.level3.keyB = "world"

    def getKey1(self):
        return 1

    def key2(self):
        return 2

    def setKey4_(self, value):
        self._key4 = value * 4

    def setKey5_(self, value):
        self.key5 = value * 5

    def keyRaisingValueError(self):
        raise ValueError("42")

    def keyRaisingNSUnknownKeyException(self):
        return self.valueForKey_("thisKeyDoesNotExist")

    def keyReturningSameSlector(self):
        return self.keyReturningSameSelector

    def keyReturningOtherSelector(self):
        return self.getKey1


class KeyValueClass1Explicit(Foundation.NSObject):
    def init(self):
        self = objc.super(KeyValueClass1Explicit, self).init()
        self._values = {}
        self._values["key3"] = 3
        self._values["key4"] = "4"
        self._values["_private"] = "private"
        return self

    def addMultiple(self):
        self._values["multiple"] = KeyValueClass1Explicit.alloc().init()
        self._values["multiple"]._values[
            "level2"
        ] = KeyValueClass1Explicit.alloc().init()
        self._values["multiple"]._values["level2"]._values[
            "level3"
        ] = KeyValueClass1Explicit.alloc().init()
        self._values["multiple"]._values["level2"]._values["level3"]._values[
            "keyA"
        ] = "hello"
        self._values["multiple"]._values["level2"]._values["level3"]._values[
            "keyB"
        ] = "world"

    def valueForKey_(self, key):
        if key == "key1":
            return 1

        elif key == "key2":
            return 2

        return self._values[key]

    def storedValueForKey_(self, key):
        return self.valueForKey_(key)

    def setValue_forKey_(self, value, key):
        if key == "key4":
            value = value * 4
        elif key == "key5":
            value = value * 5

        self._values[key] = value

    def takeStoredValue_forKey_(self, value, key):
        self.setValue_forKey_(value, key)

    def takeValue_forKey_(self, value, key):
        self.setValue_forKey_(value, key)


class KeyValueClass4(Foundation.NSObject):
    __slots__ = ("foo",)

    def init(self):
        self = objc.super(KeyValueClass4, self).init()
        self.foo = "foobar"
        return self

    # Definition for property 'bar'. Use odd names for the methods
    # because the KeyValue support recognizes the usual names.
    @objc.python_method
    def read_bar(self):
        return self.foo + self.foo

    @objc.python_method
    def write_bar(self, value):
        self.foo = value + value

    bar = property(read_bar, write_bar)

    roprop = property(lambda self: "read-only")


class KVOClass(Foundation.NSObject):
    def automaticallyNotifiesObserversForKey_(self, aKey):
        return objc.NO

    def test(self):
        return "test"


class KeyValueObserver(Foundation.NSObject):
    def init(self):
        self.observed = []
        return self

    def observeValueForKeyPath_ofObject_change_context_(
        self, keyPath, value, change, context
    ):
        self.observed.append((keyPath, value, change))


class PyKeyValueCoding(TestCase):
    def testNoPrivateVars(self):
        # Private instance variables ('anObject.__value') are not accessible using
        # key-value coding.
        o = KeyValueClass1.alloc().init()
        self.assertRaises(KeyError, STUB.keyValue_forObject_key_, 0, o, "private")

    def testValueForKey(self):
        o = KeyValueClass1.alloc().init()
        o.addMultiple()

        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key1"), 1)

        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key2"), 2)
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key3"), 3)
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key4"), "4")
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "multiple"), o.multiple)

        self.assertRaises(KeyError, STUB.keyValue_forObject_key_, 0, o, "nokey")

        self.assertRaises(
            ValueError, STUB.keyValue_forObject_key_, 0, o, "keyRaisingValueError"
        )
        self.assertRaises(
            KeyError,
            STUB.keyValue_forObject_key_,
            0,
            o,
            "keyRaisingNSUnknownKeyException",
        )
        self.assertRaises(
            KeyError, STUB.keyValue_forObject_key_, 0, o, "keyReturningSameSelector"
        )

        obj = STUB.keyValue_forObject_key_(0, o, "keyReturningOtherSelector")
        self.assertIsInstance(obj, objc.selector)
        self.assertEqual(obj.selector, b"getKey1")
        self.assertIs(obj.self, o)

    def testValueForKey2(self):
        o = KeyValueClass4.alloc().init()

        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "foo"), "foobar")
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "bar"), "foobarfoobar")
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "roprop"), "read-only")

    def testStoredValueForKey(self):
        o = KeyValueClass1.alloc().init()
        o.addMultiple()

        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key1"), 1)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key2"), 2)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key3"), 3)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key4"), "4")
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "multiple"), o.multiple)

        self.assertRaises(KeyError, STUB.keyValue_forObject_key_, 2, o, "nokey")

    def testStoredValueForKey2(self):
        o = KeyValueClass4.alloc().init()

        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "foo"), "foobar")
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "bar"), "foobarfoobar")
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "roprop"), "read-only")

    def testValueForKeyPath(self):
        o = KeyValueClass1.alloc().init()
        o.addMultiple()

        self.assertEqual(STUB.keyValue_forObject_key_(1, o, "multiple"), o.multiple)
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2"), o.multiple.level2
        )
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2.level3.keyA"),
            o.multiple.level2.level3.keyA,
        )
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2.level3.keyB"),
            o.multiple.level2.level3.keyB,
        )

        self.assertRaises(
            KeyError, STUB.keyValue_forObject_key_, 1, o, "multiple.level2.nokey"
        )

    @max_os_level("10.5")
    def testValuesForKeys(self):
        o = KeyValueClass1.alloc().init()

        self.assertEqual(
            STUB.keyValue_forObject_key_(3, o, ["key1", "key2", "key3", "key4"]),
            {"key1": 1, "key2": 2, "key3": 3, "key4": "4"},
        )

        self.assertRaises(
            KeyError, STUB.keyValue_forObject_key_, 3, o, ["key1", "key3", "nosuchkey"]
        )

    @max_os_level("10.5")
    def testTakeValueForKey(self):
        o = KeyValueClass1.alloc().init()

        self.assertEqual(o.key3, 3)
        STUB.setKeyValue_forObject_key_value_(0, o, "key3", "drie")
        self.assertEqual(o.key3, "drie")

        self.assertEqual(o._key4, "4")
        STUB.setKeyValue_forObject_key_value_(0, o, "key4", "vier")
        self.assertNotHasAttr(o, "key4")
        self.assertEqual(o._key4, "viervierviervier")

        o.key5 = 1
        STUB.setKeyValue_forObject_key_value_(0, o, "key5", "V")
        self.assertEqual(o.key5, "VVVVV")

        self.assertNotHasAttr(o, "key9")
        STUB.setKeyValue_forObject_key_value_(0, o, "key9", "IX")
        self.assertHasAttr(o, "key9")
        self.assertEqual(o.key9, "IX")

    @max_os_level("10.5")
    def testTakeValueForKey2(self):
        o = KeyValueClass4.alloc().init()

        self.assertEqual(o.foo, "foobar")
        STUB.setKeyValue_forObject_key_value_(0, o, "foo", "FOO")
        self.assertEqual(o.foo, "FOO")

        self.assertRaises(
            KeyError, STUB.setKeyValue_forObject_key_value_, 0, o, "key9", "IX"
        )

    def testTakeStoredValueForKey(self):
        o = KeyValueClass1.alloc().init()

        self.assertEqual(o.key3, 3)
        STUB.setKeyValue_forObject_key_value_(2, o, "key3", "drie")
        self.assertEqual(o.key3, "drie")

        self.assertEqual(o._key4, "4")
        STUB.setKeyValue_forObject_key_value_(2, o, "key4", "vier")
        self.assertEqual(o._key4, "viervierviervier")

        o.key5 = 1
        STUB.setKeyValue_forObject_key_value_(2, o, "key5", "V")
        self.assertEqual(o.key5, "VVVVV")

        self.assertNotHasAttr(o, "key9")
        STUB.setKeyValue_forObject_key_value_(2, o, "key9", "IX")
        self.assertHasAttr(o, "key9")
        self.assertEqual(o.key9, "IX")

    def testStoredTakeValueForKey2(self):
        o = KeyValueClass4.alloc().init()

        self.assertEqual(o.foo, "foobar")
        STUB.setKeyValue_forObject_key_value_(2, o, "foo", "FOO")
        self.assertEqual(o.foo, "FOO")

        self.assertRaises(
            KeyError, STUB.setKeyValue_forObject_key_value_, 2, o, "key9", "IX"
        )
        self.assertRaises(
            KeyError, STUB.setKeyValue_forObject_key_value_, 2, o, "roprop", "IX"
        )

    @max_os_level("10.5")
    def testTakeValuesFromDictionary(self):
        o = KeyValueClass1.alloc().init()

        self.assertEqual(o.key3, 3)
        self.assertEqual(o._key4, "4")
        o.key5 = 1
        self.assertNotHasAttr(o, "key9")

        STUB.setKeyValue_forObject_key_value_(
            3, o, None, {"key3": b"drie", "key4": b"vier", "key5": b"V", "key9": b"IX"}
        )

        self.assertEqual(o.key3, "drie")
        self.assertEqual(o._key4, "viervierviervier")
        self.assertEqual(o.key5, "VVVVV")
        self.assertHasAttr(o, "key9")
        self.assertEqual(o.key9, "IX")

    @max_os_level("10.5")
    def testTakeValuesFromDictionary2(self):
        o = KeyValueClass4.alloc().init()

        self.assertEqual(o.foo, "foobar")
        STUB.setKeyValue_forObject_key_value_(3, o, None, {"foo": "FOO"})
        self.assertEqual(o.foo, "FOO")

        self.assertRaises(
            KeyError, STUB.setKeyValue_forObject_key_value_, 3, o, None, {"key9": "IX"}
        )
        self.assertRaises(
            KeyError,
            STUB.setKeyValue_forObject_key_value_,
            3,
            o,
            None,
            {"roprop": "IX"},
        )

    @max_os_level("10.5")
    def testTakeValueForKeyPath(self):
        o = KeyValueClass1.alloc().init()
        o.addMultiple()

        self.assertEqual(o.multiple.level2.level3.keyA, "hello")
        self.assertEqual(o.multiple.level2.level3.keyB, "world")

        STUB.setKeyValue_forObject_key_value_(
            1, o, "multiple.level2.level3.keyA", "KeyAValue"
        )
        self.assertEqual(o.multiple.level2.level3.keyA, "KeyAValue")

        STUB.setKeyValue_forObject_key_value_(
            1, o, "multiple.level2.level3.keyB", 9.999
        )
        self.assertEqual(o.multiple.level2.level3.keyB, 9.999)

    if hasattr(Foundation.NSObject, "willChangeValueForKey_"):
        # NSKeyValueObserving is only available on Panther and beyond
        def testKVO1(self):
            o = KVOClass.alloc().init()
            o.addObserver_forKeyPath_options_context_(self, "test", 0, None)
            o.removeObserver_forKeyPath_(self, "test")

        def testKVO2(self):
            """
            Check if observations work for python-based keys on ObjC classes
            """
            observer = KeyValueObserver.alloc().init()
            self.assertEqual(observer.observed, [])

            o = KeyValueClass1.alloc().init()

            o.addObserver_forKeyPath_options_context_(observer, "key3", 0, 0)
            try:
                STUB.setKeyValue_forObject_key_value_(2, o, "key3", "drie")
                self.assertEqual(o.key3, "drie")

                self.assertEqual(len(observer.observed), 1)

                keyPath, value, change = observer.observed[0]
                self.assertEqual(keyPath, "key3")
                self.assertIs(value, o)
                self.assertEqual(change, {Foundation.NSKeyValueChangeKindKey: 1})

            finally:
                o.removeObserver_forKeyPath_(observer, "key3")

        def testKVO3(self):
            """
            Check if observations work for python-based keys on ObjC classes
            """
            observer = KeyValueObserver.alloc().init()
            self.assertEqual(observer.observed, [])

            o = KeyValueClass1.alloc().init()
            STUB.setKeyValue_forObject_key_value_(2, o, "key3", "three")

            o.addObserver_forKeyPath_options_context_(
                observer,
                "key3",
                Foundation.NSKeyValueObservingOptionNew
                | Foundation.NSKeyValueObservingOptionOld,
                0,
            )
            try:
                STUB.setKeyValue_forObject_key_value_(2, o, "key3", "drie")
                self.assertEqual(o.key3, "drie")

                self.assertEqual(len(observer.observed), 1)

                keyPath, value, change = observer.observed[0]
                self.assertEqual(keyPath, "key3")
                self.assertIs(value, o)
                self.assertEqual(
                    change,
                    {
                        Foundation.NSKeyValueChangeKindKey: 1,
                        Foundation.NSKeyValueChangeNewKey: "drie",
                        Foundation.NSKeyValueChangeOldKey: "three",
                    },
                )

            finally:
                o.removeObserver_forKeyPath_(observer, "key3")


class PyKeyValueCodingExplicit(TestCase):
    def testValueForKey(self):
        o = KeyValueClass1Explicit.alloc().init()
        o.addMultiple()

        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key1"), 1)

        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key2"), 2)
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key3"), 3)
        self.assertEqual(STUB.keyValue_forObject_key_(0, o, "key4"), "4")
        self.assertEqual(
            STUB.keyValue_forObject_key_(0, o, "multiple"), o._values["multiple"]
        )

        self.assertRaises(KeyError, STUB.keyValue_forObject_key_, 0, o, "nokey")

    def testStoredValueForKey(self):
        o = KeyValueClass1Explicit.alloc().init()
        o.addMultiple()

        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key1"), 1)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key2"), 2)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key3"), 3)
        self.assertEqual(STUB.keyValue_forObject_key_(2, o, "key4"), "4")
        self.assertEqual(
            STUB.keyValue_forObject_key_(2, o, "multiple"), o._values["multiple"]
        )

        self.assertRaises(KeyError, STUB.keyValue_forObject_key_, 2, o, "nokey")

    def testValueForKeyPath(self):
        o = KeyValueClass1Explicit.alloc().init()
        o.addMultiple()

        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple"), o._values["multiple"]
        )
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2"),
            o._values["multiple"]._values["level2"],
        )
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2.level3.keyA"),
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyA"],
        )
        self.assertEqual(
            STUB.keyValue_forObject_key_(1, o, "multiple.level2.level3.keyB"),
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyB"],
        )

        self.assertRaises(
            KeyError, STUB.keyValue_forObject_key_, 1, o, "multiple.level2.nokey"
        )

    @max_os_level("10.5")
    def testValuesForKeys(self):
        o = KeyValueClass1Explicit.alloc().init()

        self.assertEqual(
            STUB.keyValue_forObject_key_(3, o, ["key1", "key2", "key3", "key4"]),
            {"key1": 1, "key2": 2, "key3": 3, "key4": "4"},
        )

        self.assertRaises(
            KeyError, STUB.keyValue_forObject_key_, 3, o, ["key1", "key3", "nosuchkey"]
        )

    def testTakeValueForKey(self):
        o = KeyValueClass1Explicit.alloc().init()

        self.assertEqual(o._values["key3"], 3)
        STUB.setKeyValue_forObject_key_value_(0, o, "key3", "drie")
        self.assertEqual(o._values["key3"], "drie")

        self.assertEqual(o._values["key4"], "4")
        STUB.setKeyValue_forObject_key_value_(0, o, "key4", "vier")
        self.assertNotHasAttr(o, "key4")
        self.assertEqual(o._values["key4"], "viervierviervier")

        o._values["key5"] = 1
        STUB.setKeyValue_forObject_key_value_(0, o, "key5", "V")
        self.assertEqual(o._values["key5"], "VVVVV")

        self.assertNotHasAttr(o, "key9")
        self.assertNotIn("key9", o._values)
        STUB.setKeyValue_forObject_key_value_(0, o, "key9", "IX")
        self.assertNotHasAttr(o, "key9")
        self.assertIn("key9", o._values)
        self.assertEqual(o._values["key9"], "IX")

    def testTakeStoredValueForKey(self):
        o = KeyValueClass1Explicit.alloc().init()

        self.assertEqual(o._values["key3"], 3)
        STUB.setKeyValue_forObject_key_value_(2, o, "key3", "drie")
        self.assertEqual(o._values["key3"], "drie")

        self.assertEqual(o._values["key4"], "4")
        STUB.setKeyValue_forObject_key_value_(2, o, "key4", "vier")
        self.assertEqual(o._values["key4"], "viervierviervier")

        o.key5 = 1
        STUB.setKeyValue_forObject_key_value_(2, o, "key5", "V")
        self.assertEqual(o._values["key5"], "VVVVV")

        self.assertNotIn("key9", o._values)
        STUB.setKeyValue_forObject_key_value_(2, o, "key9", "IX")
        self.assertIn("key9", o._values)
        self.assertEqual(o._values["key9"], "IX")

    @max_os_level("10.5")
    def testTakeValuesFromDictionary(self):
        o = KeyValueClass1Explicit.alloc().init()

        self.assertEqual(o._values["key3"], 3)
        self.assertEqual(o._values["key4"], "4")
        o._values["key5"] = 1
        self.assertNotIn("key9", o._values)

        STUB.setKeyValue_forObject_key_value_(
            3, o, None, {"key3": "drie", "key4": "vier", "key5": "V", "key9": "IX"}
        )

        self.assertEqual(o._values["key3"], "drie")
        self.assertEqual(o._values["key4"], "viervierviervier")
        self.assertEqual(o._values["key5"], "VVVVV")
        self.assertEqual(o._values["key9"], "IX")

    @max_os_level("10.5")
    def testTakeValueForKeyPath(self):
        o = KeyValueClass1Explicit.alloc().init()
        o.addMultiple()

        self.assertEqual(
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyA"],
            "hello",
        )
        self.assertEqual(
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyB"],
            "world",
        )

        STUB.setKeyValue_forObject_key_value_(
            1, o, "multiple.level2.level3.keyA", "KeyAValue"
        )
        self.assertEqual(
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyA"],
            "KeyAValue",
        )

        STUB.setKeyValue_forObject_key_value_(
            1, o, "multiple.level2.level3.keyB", 9.999
        )
        self.assertEqual(
            o._values["multiple"]._values["level2"]._values["level3"]._values["keyB"],
            9.999,
        )


class TestBaseExceptions(TestCase):
    """
    Check that NSObject implementation of Key-Value coding raises the
    exception that we expect it to raise.
    """

    def testValueForKey(self):
        o = Foundation.NSObject.alloc().init()

        self.assertRaises(KeyError, o.valueForKey_, "unknownKey")

    def testStoredValueForKey(self):
        o = Foundation.NSObject.alloc().init()

        self.assertRaises(KeyError, o.storedValueForKey_, "unknownKey")

    def testTakeStoredValue(self):
        o = Foundation.NSObject.alloc().init()

        self.assertRaises(KeyError, o.takeStoredValue_forKey_, "value", "unknownKey")



import Foundation
from PyObjCTools.TestSupport import TestCase, min_sdk_level
import objc


class TestNSURLDownloadHelper(Foundation.NSObject):
    def download_willResumeWithResponse_fromByte_(self, a, b, c):
        pass

    def download_didReceiveDataOfLength_(self, a, b):
        pass

    def download_shouldDecodeSourceDataOfMIMEType_(self, a, b):
        return 1

    def download_canAuthenticateAgainstProtectionSpace_(self, a, b):
        return 1

    def downloadShouldUseCredentialStorage_(self, a):
        return 1


class TestNSURLDownload(TestCase):
    def testMethods(self):
        self.assertResultIsBOOL(
            Foundation.NSURLDownload.canResumeDownloadDecodedWithEncodingMIMEType_
        )
        self.assertArgIsBOOL(Foundation.NSURLDownload.setDestination_allowOverwrite_, 1)
        self.assertArgIsBOOL(Foundation.NSURLDownload.setDeletesFileUponFailure_, 0)
        self.assertResultIsBOOL(Foundation.NSURLDownload.deletesFileUponFailure)

    @min_sdk_level("10.7")
    def testProtocolObjects(self):
        objc.protocolNamed("NSURLDownloadDelegate")

    def testProtocols(self):
        self.assertArgHasType(
            TestNSURLDownloadHelper.download_willResumeWithResponse_fromByte_,
            2,
            objc._C_LNG_LNG,
        )
        self.assertArgHasType(
            TestNSURLDownloadHelper.download_didReceiveDataOfLength_,
            1,
            objc._C_NSUInteger,
        )
        self.assertResultIsBOOL(
            TestNSURLDownloadHelper.download_shouldDecodeSourceDataOfMIMEType_
        )
        self.assertResultIsBOOL(
            TestNSURLDownloadHelper.download_canAuthenticateAgainstProtectionSpace_
        )
        self.assertResultIsBOOL(
            TestNSURLDownloadHelper.downloadShouldUseCredentialStorage_
        )
        self.assertArgHasType(
            TestNSURLDownloadHelper.download_willResumeWithResponse_fromByte_,
            2,
            objc._C_LNG_LNG,
        )



import AppKit
from PyObjCTools.TestSupport import TestCase


class TestNSProgressIndicator(TestCase):
    def testConstants(self):
        self.assertEqual(AppKit.NSProgressIndicatorPreferredThickness, 14)
        self.assertEqual(AppKit.NSProgressIndicatorPreferredSmallThickness, 10)
        self.assertEqual(AppKit.NSProgressIndicatorPreferredLargeThickness, 18)
        self.assertEqual(AppKit.NSProgressIndicatorPreferredAquaThickness, 12)

        self.assertEqual(AppKit.NSProgressIndicatorBarStyle, 0)
        self.assertEqual(AppKit.NSProgressIndicatorSpinningStyle, 1)

        self.assertEqual(AppKit.NSProgressIndicatorStyleBar, 0)
        self.assertEqual(AppKit.NSProgressIndicatorStyleSpinning, 1)

    def testMethods(self):
        self.assertResultIsBOOL(AppKit.NSProgressIndicator.isIndeterminate)
        self.assertArgIsBOOL(AppKit.NSProgressIndicator.setIndeterminate_, 0)
        self.assertResultIsBOOL(AppKit.NSProgressIndicator.isBezeled)
        self.assertArgIsBOOL(AppKit.NSProgressIndicator.setBezeled_, 0)
        self.assertResultIsBOOL(AppKit.NSProgressIndicator.usesThreadedAnimation)
        self.assertArgIsBOOL(AppKit.NSProgressIndicator.setUsesThreadedAnimation_, 0)
        self.assertResultIsBOOL(AppKit.NSProgressIndicator.isDisplayedWhenStopped)
        self.assertArgIsBOOL(AppKit.NSProgressIndicator.setDisplayedWhenStopped_, 0)
